"""Tests for MusicBrainz API wrapper."""

from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, ClassVar

import pytest
import requests

from beets.library import Item
from beets.test.helper import PluginMixin
from beetsplug import musicbrainz
from beetsplug.musicbrainz import MusicBrainzPlugin

from .factories import musicbrainz as factories

if TYPE_CHECKING:
    from beetsplug._utils import musicbrainz as mb

_p = pytest.param


def alias_factory(**kwargs) -> mb.Alias:
    return factories.AliasFactory.build(**kwargs)


def artist_credit_factory(**kwargs) -> mb.ArtistCredit:
    return factories.ArtistCreditFactory.build(**kwargs)


def artist_relation_factory(**kwargs) -> mb.ArtistRelation:
    return factories.ArtistRelationFactory.build(**kwargs)


def label_info_factory(**kwargs) -> mb.LabelInfo:
    return factories.LabelInfoFactory.build(**kwargs)


def recording_factory(**kwargs) -> mb.Recording:
    return factories.RecordingFactory.build(**kwargs)


def track_factory(**kwargs) -> mb.Track:
    return factories.TrackFactory.build(**kwargs)


def url_relation_factory(**kwargs) -> mb.UrlRelation:
    return factories.UrlRelationFactory.build(**kwargs)


def medium_factory(**kwargs) -> mb.Medium:
    return factories.MediumFactory.build(**kwargs)


def release_factory(**kwargs) -> mb.Release:
    return factories.ReleaseFactory.build(**kwargs)


class TestUtils:
    @pytest.mark.parametrize(
        "date, expected_parts",
        [
            ("1987-03-01", (1987, 3, 1)),
            ("1987-03", (1987, 3, None)),
            ("1987", (1987, None, None)),
        ],
    )
    def test_get_date(self, date, expected_parts):
        assert musicbrainz._get_date(date) == expected_parts

    def test_single_artist(self):
        credit = [artist_credit_factory(artist__name="Artist")]

        assert MusicBrainzPlugin()._parse_artist_credits(credit) == {
            "artist": "Artist",
            "artist_id": "00000000-0000-0000-0000-000000000001",
            "artist_sort": "Artist, The",
            "artist_credit": "Artist Credit",
            "artists": ["Artist"],
            "artists_ids": ["00000000-0000-0000-0000-000000000001"],
            "artists_sort": ["Artist, The"],
            "artists_credit": ["Artist Credit"],
        }

    def test_two_artists(self):
        credit = [
            artist_credit_factory(artist__name="Artist", joinphrase=" AND "),
            artist_credit_factory(artist__name="Other Artist", artist__index=2),
        ]

        assert MusicBrainzPlugin()._parse_artist_credits(credit) == {
            "artist": "Artist AND Other Artist",
            "artist_id": "00000000-0000-0000-0000-000000000001",
            "artist_sort": "Artist, The AND Other Artist, The",
            "artist_credit": "Artist Credit AND Other Artist Credit",
            "artists": ["Artist", "Other Artist"],
            "artists_ids": [
                "00000000-0000-0000-0000-000000000001",
                "00000000-0000-0000-0000-000000000002",
            ],
            "artists_sort": ["Artist, The", "Other Artist, The"],
            "artists_credit": ["Artist Credit", "Other Artist Credit"],
        }

    @pytest.mark.parametrize(
        "languages_config, expected_alias_name",
        [
            _p([], None, id="no alias without languages"),
            _p(["en"], "Alias en", id="en primary"),
            _p(["en_GB", "en"], "Alias en_GB", id="en_GB primary"),
            _p(["en", "en_GB"], "Alias en", id="en primary over en_GB"),
            _p(["fr"], "Alias fr_P", id="fr primary"),
            _p(["pt_BR", "fr"], "Alias fr_P", id="non-primary ignored"),
        ],
    )
    def test_preferred_alias(
        self, config, languages_config, expected_alias_name
    ):
        aliases = [
            alias_factory(locale="en"),
            alias_factory(locale="en_GB"),
            alias_factory(locale="fr", primary=False),
            alias_factory(suffix="fr_P", locale="fr"),
            alias_factory(locale="pt_BR", primary=False),
        ]

        config["import"]["languages"] = languages_config

        alias = musicbrainz._preferred_alias(aliases)

        if expected_alias_name is None:
            assert not alias
        else:
            assert alias["name"] == expected_alias_name

    @pytest.mark.parametrize(
        "label_infos, expected",
        [
            _p([], {"catalognum": None, "label": None}, id="no label"),
            _p(
                [label_info_factory(label=None)],
                {"catalognum": "LAB123", "label": None},
                id="no label",
            ),
            _p(
                [label_info_factory(label__name="[no label]")],
                {"catalognum": "LAB123", "label": None},
                id="label with ignored [no label] name",
            ),
            _p(
                [label_info_factory()],
                {"catalognum": "LAB123", "label": "Label"},
                id="normal case",
            ),
        ],
    )
    def test_parse_label_info(self, label_infos, expected):
        assert MusicBrainzPlugin._parse_label_infos(label_infos) == expected


class MusicBrainzPluginTestMixin(PluginMixin):
    plugin = "musicbrainz"

    @pytest.fixture
    def plugin_config(self):
        return {}

    @pytest.fixture
    def mb(self, plugin_config):
        self.config[self.plugin].set(plugin_config)

        return musicbrainz.MusicBrainzPlugin()


class TestParseRecording(MusicBrainzPluginTestMixin):
    def test_parse_recording(self, mb):
        recording = recording_factory(
            length=None,
            disambiguation="Recording Disambiguation",
            artist_relations=[
                artist_relation_factory(
                    type="remixer",
                    artist__index=1,
                    artist__name="Recording Remixer",
                ),
                artist_relation_factory(
                    type="arranger",
                    artist__index=2,
                    artist__name="Recording Arranger",
                ),
                artist_relation_factory(
                    type="arranger",
                    artist__index=3,
                    artist__name="Another Recording Arranger",
                ),
                artist_relation_factory(type="engineer"),
            ],
            work_relations=[
                {
                    "type": "performance",
                    "work": {
                        "id": "WORK ID",
                        "title": "WORK TITLE",
                        "artist_relations": [
                            artist_relation_factory(
                                type="lyricist",
                                artist__index=4,
                                artist__name="Recording Lyricist",
                            ),
                            artist_relation_factory(
                                type="lyricist",
                                artist__index=5,
                                artist__name="Another Recording Lyricist",
                            ),
                            artist_relation_factory(
                                type="composer",
                                artist__index=6,
                                artist__name="Recording Composer",
                            ),
                            artist_relation_factory(
                                type="composer",
                                artist__index=7,
                                artist__name="Another Recording Composer",
                            ),
                            artist_relation_factory(type="mastering"),
                        ],
                    },
                }
            ],
        )

        assert mb.track_info(recording) == {
            "album": None,
            "arrangers": ["Recording Arranger", "Another Recording Arranger"],
            "arrangers_ids": [
                "00000000-0000-0000-0000-000000000002",
                "00000000-0000-0000-0000-000000000003",
            ],
            "artist": "Recording Artist",
            "artist_credit": "Recording Artist Credit",
            "artist_id": "00000000-0000-0000-0000-000000000001",
            "artist_sort": "Recording Artist, The",
            "artists": ["Recording Artist"],
            "artists_credit": ["Recording Artist Credit"],
            "artists_ids": ["00000000-0000-0000-0000-000000000001"],
            "artists_sort": ["Recording Artist, The"],
            "bpm": None,
            "composer_sort": "Recording Composer, The, Another Recording Composer, The",
            "composers": ["Recording Composer", "Another Recording Composer"],
            "composers_ids": [
                "00000000-0000-0000-0000-000000000006",
                "00000000-0000-0000-0000-000000000007",
            ],
            "data_source": "MusicBrainz",
            "data_url": "https://musicbrainz.org/recording/00000000-0000-0000-0000-000000001001",
            "disctitle": None,
            "genres": None,
            "index": None,
            "initial_key": None,
            "isrc": None,
            "length": None,
            "lyricists": ["Recording Lyricist", "Another Recording Lyricist"],
            "lyricists_ids": [
                "00000000-0000-0000-0000-000000000004",
                "00000000-0000-0000-0000-000000000005",
            ],
            "mb_workid": "WORK ID",
            "media": None,
            "medium": None,
            "medium_index": None,
            "medium_total": None,
            "release_track_id": None,
            "remixers": ["Recording Remixer"],
            "remixers_ids": ["00000000-0000-0000-0000-000000000001"],
            "title": "Recording",
            "track_alt": None,
            "track_id": "00000000-0000-0000-0000-000000001001",
            "trackdisambig": "Recording Disambiguation",
            "work": "WORK TITLE",
            "work_disambig": None,
        }


class TestParseMedia(MusicBrainzPluginTestMixin):
    def test_multiple_mediums(self, mb):
        first_medium = medium_factory(
            title="First Medium",
            position=1,
            tracks=[
                track_factory(recording__length=100000),
                track_factory(
                    position=2,
                    recording__index=2,
                    recording__length=200000,
                    recording__title="Other Recording",
                ),
            ],
        )
        second_medium = medium_factory(
            title="Second Medium",
            position=2,
            pregap=track_factory(recording__title="Pregap", position=0),
            tracks=[track_factory()],
        )
        release = release_factory(media=[first_medium, second_medium])

        d = mb.album_info(release)

        assert d.mediums == 2
        t = d.tracks
        assert len(t) == 4

        assert t[0].title == "Recording"
        assert t[0].track_id == "00000000-0000-0000-0000-000000001001"
        assert t[0].length == 100.0
        assert t[0].medium == 1
        assert t[0].medium_index == 1
        assert t[0].index == 1
        assert t[0].disctitle == "First Medium"

        assert t[1].title == "Other Recording"
        assert t[1].track_id == "00000000-0000-0000-0000-000000001002"
        assert t[1].length == 200.0
        assert t[1].medium == 1
        assert t[1].medium_index == 2
        assert t[1].index == 2
        assert t[1].disctitle == "First Medium"

        assert t[2].title == "Pregap"
        assert t[2].medium == 2
        assert t[2].medium_index == 0
        assert t[2].index == 3
        assert t[2].disctitle == "Second Medium"

        assert t[3].medium == 2
        assert t[3].medium_index == 1
        assert t[3].index == 4
        assert t[3].disctitle == "Second Medium"

    def test_track_overrides_recording(self, mb):
        release = release_factory(
            media__0__tracks=[
                track_factory(
                    length=1000.0,
                    artist_credit=[
                        artist_credit_factory(
                            artist__name="Track Artist", joinphrase=" & "
                        ),
                        artist_credit_factory(
                            artist__name="Other Track Artist", artist__index=2
                        ),
                    ],
                    recording__length=2000.0,
                    recording__artist_credit=[
                        artist_credit_factory(
                            artist__name="Recording Artist", joinphrase=" & "
                        ),
                        artist_credit_factory(
                            artist__name="Other Recording Artist",
                            artist__index=2,
                        ),
                    ],
                )
            ]
        )

        track = mb.album_info(release).tracks[0]

        assert track.length == 1.0
        assert track.artist == "Track Artist & Other Track Artist"
        assert track.artist_id == "00000000-0000-0000-0000-000000000001"
        assert (
            track.artist_sort == "Track Artist, The & Other Track Artist, The"
        )
        assert (
            track.artist_credit
            == "Track Artist Credit & Other Track Artist Credit"
        )

        assert track.artists == ["Track Artist", "Other Track Artist"]
        assert track.artists_ids == [
            "00000000-0000-0000-0000-000000000001",
            "00000000-0000-0000-0000-000000000002",
        ]
        assert track.artists_sort == [
            "Track Artist, The",
            "Other Track Artist, The",
        ]
        assert track.artists_credit == [
            "Track Artist Credit",
            "Other Track Artist Credit",
        ]

    def test_missing_tracks(self, mb):
        release = release_factory(
            media=[
                medium_factory(),
                medium_factory(
                    tracks=[
                        track_factory(),
                        track_factory(
                            recording__title="Other Recording",
                            recording__disambiguation="SECOND TRACK",
                        ),
                    ]
                ),
            ]
        )

        assert mb.album_info(release).mediums == 2

    @pytest.mark.parametrize(
        "beets_match_config, expected_titles",
        [
            _p({}, ("Audio",), id="only audio tracks by default"),
            _p(
                {"ignore_data_tracks": False},
                ("Audio", "Data"),
                id="include data tracks",
            ),
            _p(
                {"ignore_data_tracks": False, "ignore_video_tracks": False},
                ("Audio", "Video: Video", "Data"),
                id="include data and video tracks",
            ),
            _p({"ignored_media": "Vinyl"}, (), id="ignore all tracks"),
        ],
    )
    def test_data_tracks(self, config, beets_match_config, mb, expected_titles):
        medium = medium_factory(
            format="Vinyl",
            tracks=[
                track_factory(recording__title="Audio"),
                track_factory(recording__title="[data track]"),
                track_factory(recording__title="Video", recording__video=True),
            ],
            data_tracks=[track_factory(recording__title="Data")],
        )
        release = release_factory(media=[medium])

        config.set({"match": beets_match_config})

        actual_titles = tuple(t.title for t in mb.album_info(release).tracks)

        assert actual_titles == expected_titles


class TestParseRelease(MusicBrainzPluginTestMixin):
    def test_parse_release(self, config, mb):
        config["match"]["preferred"]["countries"] = ["US"]
        mb.config.set(
            {
                "external_ids": {
                    "discogs": True,
                    "bandcamp": True,
                    "spotify": False,
                }
            }
        )

        release = release_factory(
            url_relations=[
                url_relation_factory(
                    url__resource="https://discogs.com/release/123456"
                ),
                url_relation_factory(
                    url__resource="https://open.spotify.com/album/ABCDab2ImQyHZ9sXCXFyZ8"
                ),
                url_relation_factory(
                    url__resource="https://somemusic.bandcamp.com/album/somealbum"
                ),
            ]
        )
        d = mb.album_info(release)

        assert d == {
            "album": "Album",
            "album_id": "00000000-0000-0000-0000-000001000001",
            "albumdisambig": "Album Disambiguation",
            "albumstatus": "Official",
            "albumtype": "album",
            "albumtypes": ["album"],
            "artist": "Artist",
            "artist_credit": "Artist Credit",
            "artist_id": "00000000-0000-0000-0000-000000000011",
            "artist_sort": "Artist, The",
            "artists": ["Artist"],
            "artists_credit": ["Artist Credit"],
            "artists_ids": ["00000000-0000-0000-0000-000000000011"],
            "artists_sort": ["Artist, The"],
            "asin": "Album Asin",
            "bandcamp_album_id": "https://somemusic.bandcamp.com/album/somealbum",
            "barcode": "0000000000000",
            "catalognum": "LAB123",
            "country": "US",
            "data_source": "MusicBrainz",
            "data_url": "https://musicbrainz.org/release/00000000-0000-0000-0000-000001000001",
            "day": 1,
            "discogs_album_id": "123456",
            "discogs_albumid": None,
            "discogs_artistid": None,
            "discogs_labelid": None,
            "genres": None,
            "label": "Label",
            "language": "eng",
            "media": "Digital Media",
            "mediums": 1,
            "month": 1,
            "original_day": 3,
            "original_month": 2,
            "original_year": 2001,
            "release_group_title": "Release Group",
            "releasegroup_id": "00000000-0000-0000-0000-000000000101",
            "releasegroupdisambig": "Release Group Disambiguation",
            "script": "Latn",
            "style": None,
            "tracks": [
                {
                    "album": None,
                    "arrangers": None,
                    "arrangers_ids": [],
                    "artist": "Recording Artist",
                    "artist_credit": "Recording Artist Credit",
                    "artist_id": "00000000-0000-0000-0000-000000000001",
                    "artist_sort": "Recording Artist, The",
                    "artists": ["Recording Artist"],
                    "artists_credit": ["Recording Artist Credit"],
                    "artists_ids": ["00000000-0000-0000-0000-000000000001"],
                    "artists_sort": ["Recording Artist, The"],
                    "bpm": None,
                    "composer_sort": None,
                    "composers": None,
                    "composers_ids": [],
                    "data_source": "MusicBrainz",
                    "data_url": "https://musicbrainz.org/recording/00000000-0000-0000-0000-000000001001",
                    "disctitle": "Medium",
                    "genres": None,
                    "index": 1,
                    "initial_key": None,
                    "isrc": None,
                    "length": 0.36,
                    "lyricists": None,
                    "lyricists_ids": [],
                    "mb_workid": None,
                    "media": "Digital Media",
                    "medium": 1,
                    "medium_index": 1,
                    "medium_total": 1,
                    "release_track_id": "00000000-0000-0000-0000-000000010001",
                    "remixers": None,
                    "remixers_ids": [],
                    "title": "Recording",
                    "track_alt": "A1",
                    "track_id": "00000000-0000-0000-0000-000000001001",
                    "trackdisambig": None,
                    "work": None,
                    "work_disambig": None,
                }
            ],
            "va": False,
            "year": 2020,
        }

    def test_detect_various_artists(self, mb):
        release = release_factory(
            artist_credit=[
                artist_credit_factory(artist__id=musicbrainz.VARIOUS_ARTISTS_ID)
            ]
        )

        assert mb.album_info(release).va

    def test_missing_language(self, mb):
        release = release_factory(text_representation__language=None)

        assert mb.album_info(release).language is None

    @pytest.mark.parametrize(
        "plugin_config, expected_genres",
        [
            _p({"genres": False}, None, id="genres disabled"),
            _p({"genres": True, "genres_tag": "genre"}, ["Genre"], id="use genres"),
            _p({"genres": True, "genres_tag": "tag"}, ["Tag"], id="use tags"),
        ],
    )  # fmt: skip
    def test_genres(self, mb, expected_genres):
        assert mb.album_info(release_factory()).genres == expected_genres

    @pytest.mark.parametrize(
        "plugin_config", [_p({"aliases_as_credits": True})]
    )
    def test_parse_aliased_titles(self, config, mb: MusicBrainzPlugin):
        release = release_factory()

        config["import"]["languages"] = ["en"]

        d = mb.album_info(release)

        assert d.album == "Album Alias en"
        assert d.release_group_title == "Release Group Alias en"
        assert d.tracks[0].title == "Recording Alias en"

        album_artist = "Artist Alias en"
        assert d.artist == album_artist
        assert d.artists == [album_artist]
        # There is no artist credit specific alias
        assert d.artist_credit == album_artist
        assert d.artists_credit == [album_artist]

        album_artist_sort = "Artist Alias en, The"
        assert d.artist_sort == album_artist_sort
        assert d.artists_sort == [album_artist_sort]

        first_track = d.tracks[0]

        track_artist = "Recording Artist Alias en"
        assert first_track.artist == track_artist
        assert first_track.artists == [track_artist]
        # There is no artist credit specific alias
        assert first_track.artist_credit == track_artist
        assert first_track.artists_credit == [track_artist]

        track_artist_sort = "Recording Artist Alias en, The"
        assert first_track.artist_sort == track_artist_sort
        assert first_track.artists_sort == [track_artist_sort]

    @pytest.mark.parametrize(
        "plugin_config, expected_artist_credit",
        [
            _p(
                {"aliases_as_credits": False},
                {
                    "track_artist_credit": "Recording Artist Credit",
                    "track_artists_credit": ["Recording Artist Credit"],
                    "artist_credit": "Artist Credit",
                    "artists_credit": ["Artist Credit"],
                },
                id="no aliases",
            ),
            _p(
                {"aliases_as_credits": True},
                {
                    "artist_credit": "Artist Alias en",
                    "artists_credit": ["Artist Alias en"],
                    "track_artist_credit": "Recording Artist Alias en",
                    "track_artists_credit": ["Recording Artist Alias en"],
                },
                id="aliases",
            ),
        ],
    )
    def test_aliases_as_credits(
        self, config, mb: MusicBrainzPlugin, expected_artist_credit
    ):
        config["import"]["languages"] = ["en"]
        release = release_factory()

        d = mb.album_info(release)

        assert d.artist_credit == expected_artist_credit["artist_credit"]
        assert d.artists_credit == expected_artist_credit["artists_credit"]

        first_track = d.tracks[0]

        assert (
            first_track.artist_credit
            == expected_artist_credit["track_artist_credit"]
        )
        assert (
            first_track.artists_credit
            == expected_artist_credit["track_artists_credit"]
        )

    def test_ensure_complete_recordings(self, monkeypatch, mb):
        titles = ["Recording", "Other Recording"]
        initial_recordings = [
            recording_factory(index=idx, title=t)
            for idx, t in enumerate(titles)
        ]
        complete_recordings = [
            {**r, "url_relations": [url_relation_factory()]}
            for r in initial_recordings
        ]

        monkeypatch.setattr("beetsplug.musicbrainz.BROWSE_CHUNKSIZE", 1)
        monkeypatch.setattr("beetsplug.musicbrainz.BROWSE_MAXTRACKS", 1)
        monkeypatch.setattr(
            mb.mb_api,
            "browse_recordings",
            lambda offset=0, **__: [complete_recordings[offset]],
        )

        release = release_factory(
            media__0__tracks=[
                track_factory(recording=r) for r in initial_recordings
            ]
        )

        mb._ensure_complete_recordings(release)

        new_recordings = [
            t["recording"] for m in release["media"] for t in m["tracks"]
        ]
        assert new_recordings == complete_recordings

    def test_album_info_browse_recordings_prefers_alias_over_track_title(
        self, config, monkeypatch, mb
    ):
        config["import"]["languages"] = ["en"]

        recording = recording_factory(
            title="Recording Title",
            aliases=[alias_factory(type="Recording name", locale="en")],
        )
        browsed_recording = {
            **recording,
            "url_relations": [url_relation_factory()],
        }

        monkeypatch.setattr("beetsplug.musicbrainz.BROWSE_CHUNKSIZE", 1)
        monkeypatch.setattr("beetsplug.musicbrainz.BROWSE_MAXTRACKS", 0)
        monkeypatch.setattr(
            mb.mb_api,
            "browse_recordings",
            lambda offset=0, **__: [browsed_recording] if offset == 0 else [],
        )

        release = release_factory(
            media__0__tracks=[
                track_factory(recording=recording, title="Track Title")
            ]
        )

        album = mb.album_info(release)

        assert album.tracks[0].title == "Alias en"


class TestPseudoRelease(MusicBrainzPluginTestMixin):
    ACTUAL_RELEASE = release_factory(index=1, country="US")
    PSEUDO_WITHOUT_LINKS = release_factory(
        index=2, status="Pseudo-Release", country=None
    )
    PSEUDO_INVALID_LINK = release_factory(
        index=3,
        status="Pseudo-Release",
        country=None,
        release_relations=[
            {
                "type": "remaster",
                "direction": "backward",
                "release": {"id": "d2a6f856-b553-40a0-ac54-a321e8e2da01"},
            }
        ],
    )
    PSEUDO_VALID_LINK = release_factory(
        index=4,
        status="Pseudo-Release",
        country=None,
        release_relations=[
            {
                "type": "transl-tracklisting",
                "direction": "backward",
                "release": {"id": ACTUAL_RELEASE["id"]},
            }
        ],
    )

    @pytest.fixture(autouse=True)
    def setup_album_lookup(self, monkeypatch, mb):
        releases = [
            self.ACTUAL_RELEASE,
            self.PSEUDO_WITHOUT_LINKS,
            self.PSEUDO_INVALID_LINK,
            self.PSEUDO_VALID_LINK,
        ]
        release_by_id = {r["id"]: r for r in releases}

        monkeypatch.setattr(
            mb.mb_api, "get_release", lambda id_: release_by_id[id_]
        )

    @pytest.mark.parametrize(
        "release_id, expected_country",
        [
            _p(ACTUAL_RELEASE["id"], "US", id="actual release"),
            _p(PSEUDO_WITHOUT_LINKS["id"], None, id="pseudo without links"),
            _p(PSEUDO_INVALID_LINK["id"], None, id="pseudo with invalid link"),
            _p(PSEUDO_VALID_LINK["id"], "US", id="pseudo with valid link"),
        ],
    )
    def test_follow_pseudo_release(self, mb, release_id, expected_country):
        album = mb.album_for_id(release_id)

        assert album
        assert album.country == expected_country


class TestMusicBrainzPlugin(MusicBrainzPluginTestMixin):
    mbid = "d2a6f856-b553-40a0-ac54-a321e8e2da99"
    RECORDING: ClassVar[mb.Recording] = recording_factory()

    @pytest.mark.parametrize(
        "plugin_config,va_likely,expected_additional_criteria",
        [
            _p({}, False, {"artist": "Artist "}, id="default"),
            _p(
                {},
                True,
                {"arid": "89ad4ac3-39f7-470e-963a-56509c546377"},
                id="va likely",
            ),
            _p(
                {"extra_tags": ["label", "catalognum"]},
                False,
                {"artist": "Artist ", "label": "abc", "catno": "ABC123"},
                id="value-based extra_tags",
            ),
            _p(
                {"extra_tags": ["alias", "tracks"]},
                False,
                {"artist": "Artist ", "alias": " Album", "tracks": "3"},
                id="non-value-based extra_tags",
            ),
        ],
    )
    def test_get_album_criteria(
        self, mb, va_likely, expected_additional_criteria
    ):
        items = [
            Item(catalognum="ABC 123", label="abc"),
            Item(catalognum="ABC 123", label="abc"),
            Item(catalognum="ABC 123", label="def"),
        ]

        assert mb.get_album_criteria(items, "Artist ", " Album", va_likely) == {
            "release": " Album",
            **expected_additional_criteria,
        }

    def test_item_candidates(self, monkeypatch, mb):
        monkeypatch.setattr(
            "beetsplug._utils.musicbrainz.MusicBrainzAPI.get_json",
            lambda *_, **__: {"recordings": [self.RECORDING]},
        )
        monkeypatch.setattr(
            "beetsplug._utils.musicbrainz.MusicBrainzAPI.get_recording",
            lambda *_, **__: self.RECORDING,
        )

        candidates = list(mb.item_candidates(Item(), "hello", "there"))

        assert len(candidates) == 1
        assert candidates[0].track_id == self.RECORDING["id"]

    def test_candidates(self, monkeypatch, mb):
        monkeypatch.setattr(
            "beetsplug._utils.musicbrainz.MusicBrainzAPI.get_json",
            lambda *_, **__: {"releases": [{"id": self.mbid}]},
        )
        monkeypatch.setattr(
            "beetsplug._utils.musicbrainz.MusicBrainzAPI.get_release",
            lambda *_, **__: release_factory(
                id=self.mbid, media=[medium_factory()]
            ),
        )
        candidates = list(mb.candidates([], "hello", "there", False))

        assert len(candidates) == 1
        assert (
            candidates[0].tracks[0].track_id
            == "00000000-0000-0000-0000-000000001001"
        )
        assert candidates[0].album == "Album"

    def test_import_handles_404_gracefully(self, mb, requests_mock):
        id_ = uuid.uuid4()
        response = requests.Response()
        response.status_code = 404
        requests_mock.get(
            f"/ws/2/release/{id_}",
            exc=requests.exceptions.HTTPError(response=response),
        )
        res = mb.album_for_id(str(id_))
        assert res is None

    def test_import_propagates_non_404_errors(self, mb):
        class DummyResponse:
            status_code = 500

        error = requests.exceptions.HTTPError(response=DummyResponse())

        def raise_error(*args, **kwargs):
            raise error

        # Simulate mb.mb_api.get_release raising a non-404 HTTP error
        mb.mb_api.get_release = raise_error

        with pytest.raises(requests.exceptions.HTTPError) as excinfo:
            mb.album_for_id(str(uuid.uuid4()))

        # Ensure the exact error is propagated, not swallowed
        assert excinfo.value is error

    @pytest.mark.parametrize(
        "input_,expected",
        [
            ("??-??-??", (None, None, None)),
            ("??-01-??", (None, 1, None)),
            ("??-??-02", (None, None, 2)),
            ("??-01-02", (None, 1, 2)),
            ("2010-??-01", (2010, None, 1)),
            ("2010-01-not_an_int", (2010, 1, None)),
            ("2010", (2010, None, None)),
            ("2010-01", (2010, 1, None)),
            ("2010-01-02", (2010, 1, 2)),
        ],
    )
    def test_get_date(self, input_, expected):
        assert musicbrainz._get_date(input_) == expected



"""Synchronize information from music player libraries"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from importlib import import_module
from typing import TYPE_CHECKING, ClassVar

from confuse import ConfigValueError

from beets import ui
from beets.plugins import BeetsPlugin

if TYPE_CHECKING:
    from beets.dbcore import types

METASYNC_MODULE = "beetsplug.metasync"

# Dictionary to map the MODULE and the CLASS NAME of meta sources
SOURCES = {"amarok": "Amarok", "itunes": "Itunes"}


class MetaSource(metaclass=ABCMeta):
    item_types: ClassVar[dict[str, types.Type]]

    def __init__(self, config, log):
        self.config = config
        self._log = log

    @abstractmethod
    def sync_from_source(self, item):
        pass


def load_meta_sources():
    """Returns a dictionary of all the MetaSources
    E.g., {'itunes': Itunes} with isinstance(Itunes, MetaSource) true
    """
    meta_sources = {}

    for module_path, class_name in SOURCES.items():
        module = import_module(f"{METASYNC_MODULE}.{module_path}")
        meta_sources[class_name.lower()] = getattr(module, class_name)

    return meta_sources


META_SOURCES = load_meta_sources()


def load_item_types():
    """Returns a dictionary containing the item_types of all the MetaSources"""
    item_types = {}
    for meta_source in META_SOURCES.values():
        item_types.update(meta_source.item_types)
    return item_types


class MetaSyncPlugin(BeetsPlugin):
    item_types = load_item_types()

    def __init__(self):
        super().__init__()

    def commands(self):
        cmd = ui.Subcommand(
            "metasync", help="update metadata from music player libraries"
        )
        cmd.parser.add_option(
            "-p",
            "--pretend",
            action="store_true",
            help="show all changes but do nothing",
        )
        cmd.parser.add_option(
            "-s",
            "--source",
            default=[],
            action="append",
            dest="sources",
            help="comma-separated list of sources to sync",
        )
        cmd.parser.add_format_option()
        cmd.func = self.func
        return [cmd]

    def func(self, lib, opts, args):
        """Command handler for the metasync function."""
        pretend = opts.pretend

        sources = []
        for source in opts.sources:
            sources.extend(source.split(","))

        sources = sources or self.config["source"].as_str_seq()

        meta_source_instances = {}
        items = lib.items(args)

        # Avoid needlessly instantiating meta sources (can be expensive)
        if not items:
            self._log.info("No items found matching query")
            return

        # Instantiate the meta sources
        for player in sources:
            try:
                cls = META_SOURCES[player]
            except KeyError:
                self._log.error("Unknown metadata source '{}'", player)

            try:
                meta_source_instances[player] = cls(self.config, self._log)
            except (ImportError, ConfigValueError) as e:
                self._log.error(
                    "Failed to instantiate metadata source {!r}: {}", player, e
                )

        # Avoid needlessly iterating over items
        if not meta_source_instances:
            self._log.error("No valid metadata sources found")
            return

        # Sync the items with all of the meta sources
        for item in items:
            for meta_source in meta_source_instances.values():
                meta_source.sync_from_source(item)

            changed = ui.show_model_changes(item)

            if changed and not pretend:
                item.store()



"""Simple library to work out if a file is hidden on different platforms."""

from __future__ import annotations

import ctypes
import os
import stat
import sys
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from beets.util import PathLike


def is_hidden(path: PathLike) -> bool:
    """
    Determine whether the given path is treated as a 'hidden file' by the OS.
    """
    if not isinstance(path, Path):
        path = os.fsdecode(path)
    path = Path(path)

    # TODO: Avoid doing a platform check on every invocation of the function.
    # TODO: Stop supporting 'bytes' inputs once 'pathlib' is fully integrated.

    if sys.platform == "win32":
        # On Windows, we check for an FS-provided attribute.

        # FILE_ATTRIBUTE_HIDDEN = 2 (0x2) from GetFileAttributes documentation.
        hidden_mask = 2

        # Retrieve the attributes for the file.
        attrs = ctypes.windll.kernel32.GetFileAttributesW(str(path))

        # Ensure the attribute mask is valid.
        if attrs < 0:
            return False

        # Check for the hidden attribute.
        return attrs & hidden_mask

    # On OS X, we check for an FS-provided attribute.
    if sys.platform == "darwin":
        if hasattr(os.stat_result, "st_flags") and hasattr(stat, "UF_HIDDEN"):
            if path.lstat().st_flags & stat.UF_HIDDEN:
                return True

    # On all non-Windows platforms, we check for a '.'-prefixed file name.
    if path.name.startswith("."):
        return True

    return False



"""Write paths of imported files in various formats to ease later import in a
music player. Also allow printing the new file locations to stdout in case
one wants to manually add music to a player by its path.
"""

import datetime
import os
import re

from beets import config
from beets.plugins import BeetsPlugin
from beets.util import (
    FilesystemError,
    bytestring_path,
    link,
    mkdirall,
    normpath,
    syspath,
)

M3U_DEFAULT_NAME = "imported.m3u"


def _build_m3u_session_filename(basename):
    """Builds unique m3u filename by putting current date between given
    basename and file ending."""
    date = datetime.datetime.now().strftime("%Y%m%d_%Hh%M")
    basename = re.sub(r"(\.m3u|\.M3U)", "", basename)
    return normpath(
        os.path.join(
            config["importfeeds"]["dir"].as_filename(), f"{basename}_{date}.m3u"
        )
    )


def _build_m3u_filename(basename):
    """Builds unique m3u filename by appending given basename to current
    date."""
    basename = re.sub(r"[\s,/\\'\"]", "_", basename)
    date = datetime.datetime.now().strftime("%Y%m%d_%Hh%M")
    return normpath(
        os.path.join(
            config["importfeeds"]["dir"].as_filename(), f"{date}_{basename}.m3u"
        )
    )


def _write_m3u(m3u_path, items_paths):
    """Append relative paths to items into m3u file."""
    mkdirall(m3u_path)
    with open(syspath(m3u_path), "ab") as f:
        for path in items_paths:
            f.write(path + b"\n")


class ImportFeedsPlugin(BeetsPlugin):
    def __init__(self):
        super().__init__()

        self.config.add(
            {
                "formats": [],
                "m3u_name": "imported.m3u",
                "dir": None,
                "relative_to": None,
                "absolute_path": False,
            }
        )

        relative_to = self.config["relative_to"].get()
        if relative_to:
            self.config["relative_to"] = normpath(relative_to)
        else:
            self.config["relative_to"] = self.get_feeds_dir()

        self.register_listener("album_imported", self.album_imported)
        self.register_listener("item_imported", self.item_imported)
        self.register_listener("import_begin", self.import_begin)

    def get_feeds_dir(self):
        feeds_dir = self.config["dir"].get()
        if feeds_dir:
            return os.path.expanduser(bytestring_path(feeds_dir))
        return config["directory"].as_filename()

    def _record_items(self, lib, basename, items):
        """Records relative paths to the given items for each feed format"""
        feedsdir = bytestring_path(self.get_feeds_dir())
        formats = self.config["formats"].as_str_seq()
        relative_to = self.config["relative_to"].get() or self.get_feeds_dir()
        relative_to = bytestring_path(relative_to)

        paths = []
        for item in items:
            if self.config["absolute_path"]:
                paths.append(item.path)
            else:
                try:
                    relpath = os.path.relpath(item.path, relative_to)
                except ValueError:
                    # On Windows, it is sometimes not possible to construct a
                    # relative path (if the files are on different disks).
                    relpath = item.path
                paths.append(relpath)

        if "m3u" in formats:
            m3u_basename = bytestring_path(self.config["m3u_name"].as_str())
            m3u_path = os.path.join(feedsdir, m3u_basename)
            _write_m3u(m3u_path, paths)

        if "m3u_session" in formats:
            m3u_path = os.path.join(feedsdir, self.m3u_session)
            _write_m3u(m3u_path, paths)

        if "m3u_multi" in formats:
            m3u_path = _build_m3u_filename(basename)
            _write_m3u(m3u_path, paths)

        if "link" in formats:
            for path in paths:
                dest = os.path.join(feedsdir, os.path.basename(path))
                if not os.path.exists(syspath(dest)):
                    try:
                        link(path, dest)
                    except FilesystemError as exc:
                        self._log.warning(
                            "could not create symlink for {}: {}", path, exc
                        )

        if "echo" in formats:
            self._log.info("Location of imported music:")
            for path in paths:
                self._log.info("  {}", path)

    def album_imported(self, lib, album):
        self._record_items(lib, album.album, album.items())

    def item_imported(self, lib, item):
        self._record_items(lib, item.title, [item])

    def import_begin(self, session):
        formats = self.config["formats"].as_str_seq()
        if "m3u_session" in formats:
            self.m3u_session = _build_m3u_session_filename(
                self.config["m3u_name"].as_str()
            )



from typing import Any, ClassVar
from unittest.mock import ANY, Mock, call, patch

from beets import util
from beets.library import Item
from beets.test.helper import PluginTestHelper
from beetsplug.mpdstats import MPDStats, mpd_config


class TestMPDStats(PluginTestHelper):
    plugin = "mpdstats"

    def test_update_rating(self):
        item = Item(title="title", path="", id=1)
        item.add(self.lib)

        log = Mock()
        mpdstats = MPDStats(self.lib, log)

        assert not mpdstats.update_rating(item, True)
        assert not mpdstats.update_rating(None, True)

    def test_get_item(self):
        item_path = util.normpath("/foo/bar.flac")
        item = Item(title="title", path=item_path, id=1)
        item.add(self.lib)

        log = Mock()
        mpdstats = MPDStats(self.lib, log)

        assert str(mpdstats.get_item(item_path)) == str(item)
        assert mpdstats.get_item("/some/non-existing/path") is None
        assert "item not found:" in log.info.call_args[0][0]

    STATUSES: ClassVar[list[dict[str, Any]]] = [
        {"state": "some-unknown-one"},
        {"state": "pause"},
        {"state": "play", "songid": 1, "time": "0:1"},
        {"state": "stop"},
    ]
    EVENTS = [["player"]] * (len(STATUSES) - 1) + [KeyboardInterrupt]
    item_path = util.normpath("/foo/bar.flac")
    songid = 1

    @patch(
        "beetsplug.mpdstats.MPDClientWrapper",
        return_value=Mock(
            **{
                "events.side_effect": EVENTS,
                "status.side_effect": STATUSES,
                "currentsong.return_value": (item_path, songid),
            }
        ),
    )
    def test_run_mpdstats(self, mpd_mock):
        item = Item(title="title", path=self.item_path, id=1)
        item.add(self.lib)

        log = Mock()
        try:
            MPDStats(self.lib, log).run()
        except KeyboardInterrupt:
            pass

        log.debug.assert_has_calls([call('unhandled status "{}"', ANY)])
        log.info.assert_has_calls(
            [call("pause"), call("playing {}", ANY), call("stop")]
        )

    @patch("beetsplug.mpdstats.MPDStats.run")
    def test_cli_options_override_config(self, run_mock):
        self.run_command(
            "mpdstats",
            "--host",
            "somehost",
            "--port",
            "5000",
            "--password",
            "secret",
        )

        assert mpd_config["host"].as_str() == "somehost"
        assert mpd_config["port"].get(int) == 5000
        assert mpd_config["password"].as_str() == "secret"
        run_mock.assert_called_once_with()



"""Test file manipulation functionality of Item."""

import os
import shutil
import stat
import unittest
from os.path import join
from pathlib import Path

import beets.library
from beets import util
from beets.test import _common
from beets.test._common import item, touch
from beets.test.helper import NEEDS_REFLINK, BeetsTestCase
from beets.util import MoveOperation, syspath


class MoveTest(BeetsTestCase):
    def setUp(self):
        super().setUp()

        # make a temporary file
        self.temp_music_file_name = "temp.mp3"
        self.path = self.temp_path / self.temp_music_file_name
        shutil.copy(self.resource_path, self.path)

        # add it to a temporary library
        self.i = beets.library.Item.from_path(self.path)
        self.lib.add(self.i)

        # set up the destination
        self.lib.path_formats = [
            ("default", join("$artist", "$album", "$title"))
        ]
        self.i.artist = "one"
        self.i.album = "two"
        self.i.title = "three"
        self.dest = self.lib_path / "one" / "two" / "three.mp3"

        self.otherdir = self.temp_path / "testotherdir"

    def test_move_arrives(self):
        self.i.move()
        assert self.dest.exists()

    def test_move_to_custom_dir(self):
        self.i.move(basedir=os.fsencode(self.otherdir))
        assert (self.otherdir / "one" / "two" / "three.mp3").exists()

    def test_move_departs(self):
        self.i.move()
        assert not self.path.exists()

    def test_move_in_lib_prunes_empty_dir(self):
        self.i.move()
        old_path = self.i.filepath
        assert old_path.exists()

        self.i.artist = "newArtist"
        self.i.move()
        assert not old_path.exists()
        assert not old_path.parent.exists()

    def test_copy_arrives(self):
        self.i.move(operation=MoveOperation.COPY)
        assert self.dest.exists()

    def test_copy_does_not_depart(self):
        self.i.move(operation=MoveOperation.COPY)
        assert self.path.exists()

    def test_reflink_arrives(self):
        self.i.move(operation=MoveOperation.REFLINK_AUTO)
        assert self.dest.exists()

    def test_reflink_does_not_depart(self):
        self.i.move(operation=MoveOperation.REFLINK_AUTO)
        assert self.path.exists()

    @NEEDS_REFLINK
    def test_force_reflink_arrives(self):
        self.i.move(operation=MoveOperation.REFLINK)
        assert self.dest.exists()

    @NEEDS_REFLINK
    def test_force_reflink_does_not_depart(self):
        self.i.move(operation=MoveOperation.REFLINK)
        assert self.path.exists()

    def test_move_changes_path(self):
        self.i.move()
        assert self.i.path == util.normpath(self.dest)

    def test_copy_already_at_destination(self):
        self.i.move()
        old_path = self.i.path
        self.i.move(operation=MoveOperation.COPY)
        assert self.i.path == old_path

    def test_move_already_at_destination(self):
        self.i.move()
        old_path = self.i.path
        self.i.move()
        assert self.i.path == old_path

    def test_move_file_with_colon(self):
        self.i.artist = "C:DOS"
        self.i.move()
        assert "C_DOS" in self.i.path.decode()

    def test_move_file_with_multiple_colons(self):
        # print(beets.config["replace"])
        self.i.artist = "COM:DOS"
        self.i.move()
        assert "COM_DOS" in self.i.path.decode()

    def test_move_file_with_colon_alt_separator(self):
        old = beets.config["drive_sep_replace"]
        beets.config["drive_sep_replace"] = "0"
        self.i.artist = "C:DOS"
        self.i.move()
        assert "C0DOS" in self.i.path.decode()
        beets.config["drive_sep_replace"] = old

    def test_read_only_file_copied_writable(self):
        # Make the source file read-only.
        os.chmod(syspath(self.path), 0o444)

        try:
            self.i.move(operation=MoveOperation.COPY)
            assert os.access(syspath(self.i.path), os.W_OK)
        finally:
            # Make everything writable so it can be cleaned up.
            os.chmod(syspath(self.path), 0o777)
            os.chmod(syspath(self.i.path), 0o777)

    def test_move_avoids_collision_with_existing_file(self):
        # Make a conflicting file at the destination.
        dest = self.i.destination()
        os.makedirs(syspath(os.path.dirname(dest)))
        touch(dest)

        self.i.move()
        assert self.i.path != dest
        assert os.path.dirname(self.i.path) == os.path.dirname(dest)

    @unittest.skipUnless(_common.HAVE_SYMLINK, "need symlinks")
    def test_link_arrives(self):
        self.i.move(operation=MoveOperation.LINK)
        assert self.dest.exists()
        assert os.path.islink(syspath(self.dest))
        assert self.dest.resolve() == self.path.resolve()

    @unittest.skipUnless(_common.HAVE_SYMLINK, "need symlinks")
    def test_link_does_not_depart(self):
        self.i.move(operation=MoveOperation.LINK)
        assert self.path.exists()

    @unittest.skipUnless(_common.HAVE_SYMLINK, "need symlinks")
    def test_link_changes_path(self):
        self.i.move(operation=MoveOperation.LINK)
        assert self.i.path == util.normpath(self.dest)

    @unittest.skipUnless(_common.HAVE_HARDLINK, "need hardlinks")
    def test_hardlink_arrives(self):
        self.i.move(operation=MoveOperation.HARDLINK)
        assert self.dest.exists()
        s1 = os.stat(syspath(self.path))
        s2 = os.stat(syspath(self.dest))
        assert (s1[stat.ST_INO], s1[stat.ST_DEV]) == (
            s2[stat.ST_INO],
            s2[stat.ST_DEV],
        )

    @unittest.skipUnless(_common.HAVE_HARDLINK, "need hardlinks")
    def test_hardlink_does_not_depart(self):
        self.i.move(operation=MoveOperation.HARDLINK)
        assert self.path.exists()

    @unittest.skipUnless(_common.HAVE_HARDLINK, "need hardlinks")
    def test_hardlink_changes_path(self):
        self.i.move(operation=MoveOperation.HARDLINK)
        assert self.i.path == util.normpath(self.dest)

    @unittest.skipUnless(_common.HAVE_HARDLINK, "need hardlinks")
    def test_hardlink_from_symlink(self):
        link_path = self.temp_path / "temp_link.mp3"
        link_path.symlink_to(self.temp_music_file_name)
        self.i.path = link_path
        self.i.move(operation=MoveOperation.HARDLINK)

        s1 = self.path.stat()
        s2 = self.dest.stat()
        assert (s1[stat.ST_INO], s1[stat.ST_DEV]) == (
            s2[stat.ST_INO],
            s2[stat.ST_DEV],
        )


class AlbumFileTest(BeetsTestCase):
    def setUp(self):
        super().setUp()

        # Make library and item.
        self.lib.path_formats = [
            ("default", join("$albumartist", "$album", "$title"))
        ]
        self.i = item(self.lib)
        # Make a file for the item.
        self.i.path = self.i.destination()
        util.mkdirall(self.i.path)
        touch(self.i.path)
        # Make an album.
        self.ai = self.lib.add_album((self.i,))
        # Alternate destination dir.
        self.otherdir = os.fsencode(self.temp_path / "testotherdir")

    def test_albuminfo_move_changes_paths(self):
        self.ai.album = "newAlbumName"
        self.ai.move()
        self.ai.store()
        self.i.load()

        assert b"newAlbumName" in self.i.path

    def test_albuminfo_move_moves_file(self):
        oldpath = self.i.filepath
        self.ai.album = "newAlbumName"
        self.ai.move()
        self.ai.store()
        self.i.load()

        assert not oldpath.exists()
        assert self.i.filepath.exists()

    def test_albuminfo_move_copies_file(self):
        oldpath = self.i.filepath
        self.ai.album = "newAlbumName"
        self.ai.move(operation=MoveOperation.COPY)
        self.ai.store()
        self.i.load()

        assert oldpath.exists()
        assert self.i.filepath.exists()

    @NEEDS_REFLINK
    def test_albuminfo_move_reflinks_file(self):
        oldpath = self.i.path
        self.ai.album = "newAlbumName"
        self.ai.move(operation=MoveOperation.REFLINK)
        self.ai.store()
        self.i.load()

        assert os.path.exists(oldpath)
        assert os.path.exists(self.i.path)

    def test_albuminfo_move_to_custom_dir(self):
        self.ai.move(basedir=self.otherdir)
        self.i.load()
        self.ai.store()
        assert b"testotherdir" in self.i.path


class ArtFileTest(BeetsTestCase):
    def setUp(self):
        super().setUp()

        # Make library and item.
        self.i = item(self.lib)
        self.i.path = self.i.destination()
        # Make a music file.
        util.mkdirall(self.i.path)
        touch(self.i.path)
        # Make an album.
        self.ai = self.lib.add_album((self.i,))
        # Make an art file too.
        art_bytes = self.lib.get_album(self.i).art_destination("something.jpg")
        self.art = Path(os.fsdecode(art_bytes))
        self.art.touch()
        self.ai.artpath = art_bytes
        self.ai.store()
        # Alternate destination dir.
        self.otherdir = os.fsencode(self.temp_path / "testotherdir")

    def test_art_deleted_when_items_deleted(self):
        assert self.art.exists()
        self.ai.remove(True)
        assert not self.art.exists()

    def test_art_moves_with_album(self):
        assert self.art.exists()
        oldpath = self.i.path
        self.ai.album = "newAlbum"
        self.ai.move()
        self.i.load()

        assert self.i.path != oldpath
        assert not self.art.exists()
        newart = self.lib.get_album(self.i).art_destination(self.art)
        assert Path(os.fsdecode(newart)).exists()

    def test_art_moves_with_album_to_custom_dir(self):
        # Move the album to another directory.
        self.ai.move(basedir=self.otherdir)
        self.ai.store()
        self.i.load()

        # Art should be in new directory.
        assert not self.art.exists()
        newart = self.lib.get_album(self.i).art_filepath
        assert newart.exists()
        assert "testotherdir" in str(newart)

    def test_setart_copies_image(self):
        util.remove(self.art)

        newart = self.lib_path / "newart.jpg"
        newart.touch()
        i2 = item()
        i2.path = self.i.path
        i2.artist = "someArtist"
        ai = self.lib.add_album((i2,))
        i2.move(operation=MoveOperation.COPY)

        assert ai.artpath is None
        ai.set_art(newart)
        assert ai.art_filepath.exists()

    def test_setart_to_existing_art_works(self):
        util.remove(self.art)

        # Original art.
        newart = self.lib_path / "newart.jpg"
        newart.touch()
        i2 = item()
        i2.path = self.i.path
        i2.artist = "someArtist"
        ai = self.lib.add_album((i2,))
        i2.move(operation=MoveOperation.COPY)
        ai.set_art(newart)

        # Set the art again.
        ai.set_art(ai.artpath)
        assert ai.art_filepath.exists()

    def test_setart_to_existing_but_unset_art_works(self):
        newart = self.lib_path / "newart.jpg"
        newart.touch()
        i2 = item()
        i2.path = self.i.path
        i2.artist = "someArtist"
        ai = self.lib.add_album((i2,))
        i2.move(operation=MoveOperation.COPY)

        # Copy the art to the destination.
        artdest = ai.art_destination(newart)
        shutil.copy(syspath(newart), syspath(artdest))

        # Set the art again.
        ai.set_art(artdest)
        assert ai.art_filepath.exists()

    def test_setart_to_conflicting_file_replaces_it(self):
        newart = self.lib_path / "newart.jpg"
        newart.touch()
        i2 = item()
        i2.path = self.i.path
        i2.artist = "someArtist"
        ai = self.lib.add_album((i2,))
        i2.move(operation=MoveOperation.COPY)

        # Make a file at the destination.
        artdest = ai.art_destination(newart)
        touch(artdest)

        # Set the art - should replace the existing file, not create a suffixed
        # duplicate like cover.2.jpg.
        ai.set_art(newart)
        assert artdest == ai.artpath

    def test_setart_replaces_old_art_at_different_path(self):
        newart = self.lib_path / "newart.png"
        newart.touch()
        i2 = item()
        i2.path = self.i.path
        i2.artist = "someArtist"
        ai = self.lib.add_album((i2,))
        i2.move(operation=MoveOperation.COPY)

        # Set initial art.
        ai.set_art(newart)
        old_artpath = ai.artpath
        assert os.path.exists(syspath(old_artpath))

        # Set new art with a different extension.
        another_art = self.lib_path / "another.jpg"
        another_art.touch()
        ai.set_art(another_art)

        # Old art should be removed.
        assert not os.path.exists(syspath(old_artpath))
        assert ai.art_filepath.exists()

    def test_setart_sets_permissions(self):
        util.remove(self.art)

        newart = self.lib_path / "newart.jpg"
        newart.touch()
        newart.chmod(0o400)  # read-only

        try:
            i2 = item()
            i2.path = self.i.path
            i2.artist = "someArtist"
            ai = self.lib.add_album((i2,))
            i2.move(operation=MoveOperation.COPY)
            ai.set_art(newart)

            mode = stat.S_IMODE(os.stat(syspath(ai.artpath)).st_mode)
            assert mode & stat.S_IRGRP
            assert os.access(syspath(ai.artpath), os.W_OK)

        finally:
            # Make everything writable so it can be cleaned up.
            os.chmod(syspath(newart), 0o777)
            os.chmod(syspath(ai.artpath), 0o777)

    def test_move_last_file_moves_albumart(self):
        oldartpath = self.lib.albums()[0].art_filepath
        assert oldartpath.exists()

        self.ai.album = "different_album"
        self.ai.store()
        self.ai.items()[0].move()

        artpath = self.lib.albums()[0].art_filepath
        assert "different_album" in str(artpath)
        assert artpath.exists()
        assert not oldartpath.exists()

    def test_move_not_last_file_does_not_move_albumart(self):
        i2 = item()
        i2.albumid = self.ai.id
        self.lib.add(i2)

        oldartpath = self.lib.albums()[0].art_filepath
        assert oldartpath.exists()

        self.i.album = "different_album"
        self.i.album_id = None  # detach from album
        self.i.move()

        artpath = self.lib.albums()[0].art_filepath
        assert "different_album" not in str(artpath)
        assert artpath == oldartpath
        assert oldartpath.exists()


class RemoveTest(BeetsTestCase):
    def setUp(self):
        super().setUp()

        # Make library and item.
        self.i = item(self.lib)
        self.i.path = self.i.destination()
        # Make a music file.
        util.mkdirall(self.i.path)
        touch(self.i.path)
        # Make an album with the item.
        self.ai = self.lib.add_album((self.i,))

    def test_removing_last_item_prunes_empty_dir(self):
        assert self.i.filepath.parent.exists()
        self.i.remove(True)
        assert not self.i.filepath.parent.exists()

    def test_removing_last_item_preserves_nonempty_dir(self):
        (self.i.filepath.parent / "dummy.txt").touch()
        self.i.remove(True)
        assert self.i.filepath.parent.exists()

    def test_removing_last_item_prunes_dir_with_blacklisted_file(self):
        (self.i.filepath.parent / ".DS_Store").touch()
        self.i.remove(True)
        assert not self.i.filepath.parent.exists()

    def test_removing_without_delete_leaves_file(self):
        self.i.remove(False)
        assert self.i.filepath.parent.exists()

    def test_removing_last_item_preserves_library_dir(self):
        self.i.remove(True)
        assert self.lib_path.exists()

    def test_removing_last_item_in_album_with_albumart_prunes_dir(self):
        artfile = self.temp_path / "testart.jpg"
        artfile.touch()
        self.ai.set_art(artfile)
        self.ai.store()

        self.i.remove(True)
        assert not self.i.filepath.parent.exists()



"""Test for dbcore's date-based queries."""

import time
import unittest
from datetime import datetime, timedelta

import pytest

from beets.dbcore.query import (
    DateInterval,
    DateQuery,
    InvalidQueryArgumentValueError,
    _parse_periods,
)
from beets.test.helper import ItemInDBTestCase


class TestDateInterval:
    now = datetime.now().replace(microsecond=0, second=0).isoformat()

    @pytest.mark.parametrize(
        "pattern, datestr, include",
        [
            # year precision
            ("2000..2001", "2000-01-01T00:00:00", True),
            ("2000..2001", "2001-06-20T14:15:16", True),
            ("2000..2001", "2001-12-31T23:59:59", True),
            ("2000..2001", "1999-12-31T23:59:59", False),
            ("2000..2001", "2002-01-01T00:00:00", False),
            ("2000..", "2000-01-01T00:00:00", True),
            ("2000..", "2099-10-11T00:00:00", True),
            ("2000..", "1999-12-31T23:59:59", False),
            ("..2001", "2001-12-31T23:59:59", True),
            ("..2001", "2002-01-01T00:00:00", False),
            ("-1d..1d", now, True),
            ("-2d..-1d", now, False),
            # month precision
            ("2000-06-20..2000-06-20", "2000-06-20T00:00:00", True),
            ("2000-06-20..2000-06-20", "2000-06-20T10:20:30", True),
            ("2000-06-20..2000-06-20", "2000-06-20T23:59:59", True),
            ("2000-06-20..2000-06-20", "2000-06-19T23:59:59", False),
            ("2000-06-20..2000-06-20", "2000-06-21T00:00:00", False),
            # day precision
            ("1999-12..2000-02", "1999-12-01T00:00:00", True),
            ("1999-12..2000-02", "2000-02-15T05:06:07", True),
            ("1999-12..2000-02", "2000-02-29T23:59:59", True),
            ("1999-12..2000-02", "1999-11-30T23:59:59", False),
            ("1999-12..2000-02", "2000-03-01T00:00:00", False),
            # hour precision with 'T' separator
            ("2000-01-01T12..2000-01-01T13", "2000-01-01T11:59:59", False),
            ("2000-01-01T12..2000-01-01T13", "2000-01-01T12:00:00", True),
            ("2000-01-01T12..2000-01-01T13", "2000-01-01T12:30:00", True),
            ("2000-01-01T12..2000-01-01T13", "2000-01-01T13:30:00", True),
            ("2000-01-01T12..2000-01-01T13", "2000-01-01T13:59:59", True),
            ("2000-01-01T12..2000-01-01T13", "2000-01-01T14:00:00", False),
            ("2000-01-01T12..2000-01-01T13", "2000-01-01T14:30:00", False),
            # hour precision non-range query
            ("2008-12-01T22", "2008-12-01T22:30:00", True),
            ("2008-12-01T22", "2008-12-01T23:30:00", False),
            # minute precision
            ("2000-01-01T12:30..2000-01-01T12:31", "2000-01-01T12:29:59", False),
            ("2000-01-01T12:30..2000-01-01T12:31", "2000-01-01T12:30:00", True),
            ("2000-01-01T12:30..2000-01-01T12:31", "2000-01-01T12:30:30", True),
            ("2000-01-01T12:30..2000-01-01T12:31", "2000-01-01T12:31:59", True),
            ("2000-01-01T12:30..2000-01-01T12:31", "2000-01-01T12:32:00", False),
            # second precision
            ("2000-01-01T12:30:50..2000-01-01T12:30:55", "2000-01-01T12:30:49", False),
            ("2000-01-01T12:30:50..2000-01-01T12:30:55", "2000-01-01T12:30:50", True),
            ("2000-01-01T12:30:50..2000-01-01T12:30:55", "2000-01-01T12:30:55", True),
            ("2000-01-01T12:30:50..2000-01-01T12:30:55", "2000-01-01T12:30:56", False), # unbounded  # noqa: E501
            ("..", datetime.max.isoformat(), True),
            ("..", datetime.min.isoformat(), True),
            ("..", "1000-01-01T00:00:00", True),
        ],
    )  # fmt: skip
    def test_intervals(self, pattern, datestr, include):
        (start, end) = _parse_periods(pattern)
        interval = DateInterval.from_periods(start, end)
        assert interval.contains(datetime.fromisoformat(datestr)) == include


def _parsetime(s):
    return time.mktime(datetime.strptime(s, "%Y-%m-%d %H:%M").timetuple())


class DateQueryTest(ItemInDBTestCase):
    def setUp(self):
        super().setUp()
        self.i.added = _parsetime("2013-03-30 22:21")
        self.i.store()

    def test_single_month_match_fast(self):
        query = DateQuery("added", "2013-03")
        matched = self.lib.items(query)
        assert len(matched) == 1

    def test_single_month_nonmatch_fast(self):
        query = DateQuery("added", "2013-04")
        matched = self.lib.items(query)
        assert len(matched) == 0

    def test_single_month_match_slow(self):
        query = DateQuery("added", "2013-03")
        assert query.match(self.i)

    def test_single_month_nonmatch_slow(self):
        query = DateQuery("added", "2013-04")
        assert not query.match(self.i)

    def test_single_day_match_fast(self):
        query = DateQuery("added", "2013-03-30")
        matched = self.lib.items(query)
        assert len(matched) == 1

    def test_single_day_nonmatch_fast(self):
        query = DateQuery("added", "2013-03-31")
        matched = self.lib.items(query)
        assert len(matched) == 0


class DateQueryTestRelative(ItemInDBTestCase):
    def setUp(self):
        super().setUp()

        # We pick a date near a month changeover, which can reveal some time
        # zone bugs.
        self._now = datetime(2017, 12, 31, 22, 55, 4, 101332)

        self.i.added = _parsetime(self._now.strftime("%Y-%m-%d %H:%M"))
        self.i.store()

    def test_single_month_match_fast(self):
        query = DateQuery("added", self._now.strftime("%Y-%m"))
        matched = self.lib.items(query)
        assert len(matched) == 1

    def test_single_month_nonmatch_fast(self):
        query = DateQuery(
            "added", (self._now + timedelta(days=30)).strftime("%Y-%m")
        )
        matched = self.lib.items(query)
        assert len(matched) == 0

    def test_single_month_match_slow(self):
        query = DateQuery("added", self._now.strftime("%Y-%m"))
        assert query.match(self.i)

    def test_single_month_nonmatch_slow(self):
        query = DateQuery(
            "added", (self._now + timedelta(days=30)).strftime("%Y-%m")
        )
        assert not query.match(self.i)

    def test_single_day_match_fast(self):
        query = DateQuery("added", self._now.strftime("%Y-%m-%d"))
        matched = self.lib.items(query)
        assert len(matched) == 1

    def test_single_day_nonmatch_fast(self):
        query = DateQuery(
            "added", (self._now + timedelta(days=1)).strftime("%Y-%m-%d")
        )
        matched = self.lib.items(query)
        assert len(matched) == 0


class DateQueryTestRelativeMore(ItemInDBTestCase):
    def setUp(self):
        super().setUp()
        self.i.added = _parsetime(datetime.now().strftime("%Y-%m-%d %H:%M"))
        self.i.store()

    def test_relative(self):
        for timespan in ["d", "w", "m", "y"]:
            query = DateQuery("added", f"-4{timespan}..+4{timespan}")
            matched = self.lib.items(query)
            assert len(matched) == 1

    def test_relative_fail(self):
        for timespan in ["d", "w", "m", "y"]:
            query = DateQuery("added", f"-2{timespan}..-1{timespan}")
            matched = self.lib.items(query)
            assert len(matched) == 0

    def test_start_relative(self):
        for timespan in ["d", "w", "m", "y"]:
            query = DateQuery("added", f"-4{timespan}..")
            matched = self.lib.items(query)
            assert len(matched) == 1

    def test_start_relative_fail(self):
        for timespan in ["d", "w", "m", "y"]:
            query = DateQuery("added", f"4{timespan}..")
            matched = self.lib.items(query)
            assert len(matched) == 0

    def test_end_relative(self):
        for timespan in ["d", "w", "m", "y"]:
            query = DateQuery("added", f"..+4{timespan}")
            matched = self.lib.items(query)
            assert len(matched) == 1

    def test_end_relative_fail(self):
        for timespan in ["d", "w", "m", "y"]:
            query = DateQuery("added", f"..-4{timespan}")
            matched = self.lib.items(query)
            assert len(matched) == 0


class DateQueryConstructTest(unittest.TestCase):
    def test_long_numbers(self):
        with pytest.raises(InvalidQueryArgumentValueError):
            DateQuery("added", "1409830085..1412422089")

    def test_too_many_components(self):
        with pytest.raises(InvalidQueryArgumentValueError):
            DateQuery("added", "12-34-56-78")

    def test_invalid_date_query(self):
        q_list = [
            "2001-01-0a",
            "2001-0a",
            "200a",
            "2001-01-01..2001-01-0a",
            "2001-0a..2001-01",
            "200a..2002",
            "20aa..",
            "..2aa",
        ]
        for q in q_list:
            with pytest.raises(InvalidQueryArgumentValueError):
                DateQuery("added", q)

    def test_pipe_in_relative_date_query(self):
        # A stray "|" (e.g. a user expecting OR syntax like "added:2000|2001")
        # must raise a clean InvalidQueryArgumentValueError rather than an
        # uncaught KeyError from the relative-date parser.
        for q in ["2000|2001", "1|", "5|3d"]:
            with pytest.raises(InvalidQueryArgumentValueError):
                DateQuery("added", q)

    def test_datetime_uppercase_t_separator(self):
        date_query = DateQuery("added", "2000-01-01T12")
        assert date_query.interval.start == datetime(2000, 1, 1, 12)
        assert date_query.interval.end == datetime(2000, 1, 1, 13)

    def test_datetime_lowercase_t_separator(self):
        date_query = DateQuery("added", "2000-01-01t12")
        assert date_query.interval.start == datetime(2000, 1, 1, 12)
        assert date_query.interval.end == datetime(2000, 1, 1, 13)

    def test_datetime_space_separator(self):
        date_query = DateQuery("added", "2000-01-01 12")
        assert date_query.interval.start == datetime(2000, 1, 1, 12)
        assert date_query.interval.end == datetime(2000, 1, 1, 13)

    def test_datetime_invalid_separator(self):
        with pytest.raises(InvalidQueryArgumentValueError):
            DateQuery("added", "2000-01-01x12")



#!/usr/bin/env python3

"""A tiny tool used to test the `convert` plugin. It copies a file and appends
a specified text tag.
"""

import sys


def convert(in_file, out_file, tag):
    """Copy `in_file` to `out_file` and append the string `tag`."""
    if not isinstance(tag, bytes):
        tag = tag.encode("utf-8")

    with open(out_file, "wb") as out_f:
        with open(in_file, "rb") as in_f:
            out_f.write(in_f.read())
        out_f.write(tag)


if __name__ == "__main__":
    convert(sys.argv[1], sys.argv[2], sys.argv[3])



"""The `update` command: Update library contents according to on-disk tags."""

import os

from beets import library, logging, ui
from beets.util import ancestry, syspath
from beets.util.color import colorize

from .utils import do_query

# Global logger.
log = logging.getLogger("beets")


def update_items(lib, query, album, move, pretend, fields, exclude_fields=None):
    """For all the items matched by the query, update the library to
    reflect the item's embedded tags.
    :param fields: The fields to be stored. If not specified, all fields will
    be.
    :param exclude_fields: The fields to not be stored. If not specified, all
    fields will be.
    """
    with lib.transaction():
        items, _ = do_query(lib, query, album)
        if move and fields is not None and "path" not in fields:
            # Special case: if an item needs to be moved, the path field has to
            # updated; otherwise the new path will not be reflected in the
            # database.
            fields.append("path")
        if fields is None:
            # no fields were provided, update all media fields
            item_fields = fields or library.Item._media_fields
            if move and "path" not in item_fields:
                # move is enabled, add 'path' to the list of fields to update
                item_fields.add("path")
        else:
            # fields was provided, just update those
            item_fields = fields
        # get all the album fields to update
        album_fields = fields or library.Album._fields.keys()
        if exclude_fields:
            # remove any excluded fields from the item and album sets
            item_fields = [f for f in item_fields if f not in exclude_fields]
            album_fields = [f for f in album_fields if f not in exclude_fields]

        # Walk through the items and pick up their changes.
        affected_albums = set()
        for item in items:
            # Item deleted?
            if not item.path or not os.path.exists(syspath(item.path)):
                ui.print_(format(item))
                ui.print_(colorize("text_error", "  deleted"))
                if not pretend:
                    item.remove(True)
                affected_albums.add(item.album_id)
                continue

            # Did the item change since last checked?
            if item.current_mtime() <= item.mtime:
                log.debug(
                    "skipping {0.filepath} because mtime is up to date ({0.mtime})",
                    item,
                )
                continue

            # Read new data.
            try:
                item.read()
            except library.ReadError as exc:
                log.error("error reading {.filepath}: {}", item, exc)
                continue

            # Special-case album artist when it matches track artist. (Hacky
            # but necessary for preserving album-level metadata for non-
            # autotagged imports.)
            if not item.albumartist:
                old_item = lib.get_item(item.id)
                if old_item.albumartist == old_item.artist == item.artist:
                    item.albumartist = old_item.albumartist
                    item._dirty.discard("albumartist")

            # Check for and display changes.
            changed = ui.show_model_changes(item, fields=item_fields)

            # Save changes.
            if not pretend:
                if changed:
                    # Move the item if it's in the library.
                    if move and lib.directory in ancestry(item.path):
                        item.move(store=False)

                    item.store(fields=item_fields)
                    affected_albums.add(item.album_id)
                else:
                    # The file's mtime was different, but there were no
                    # changes to the metadata. Store the new mtime,
                    # which is set in the call to read(), so we don't
                    # check this again in the future.
                    item.store(fields=item_fields)

        # Skip album changes while pretending.
        if pretend:
            return

        # Modify affected albums to reflect changes in their items.
        for album_id in affected_albums:
            if album_id is None:  # Singletons.
                continue
            album = lib.get_album(album_id)
            if not album:  # Empty albums have already been removed.
                log.debug("emptied album {}", album_id)
                continue
            first_item = album.items().get()

            # Update album structure to reflect an item in it.
            for key in library.Album.item_keys:
                album[key] = first_item[key]
            album.store(fields=album_fields)

            # Move album art (and any inconsistent items).
            if move and lib.directory in ancestry(first_item.path):
                log.debug("moving album {}", album_id)

                # Manually moving and storing the album.
                items = list(album.items())
                for item in items:
                    item.move(store=False, with_album=False)
                    item.store(fields=item_fields)
                album.move(store=False)
                album.store(fields=album_fields)


def update_func(lib, opts, args):
    # Verify that the library folder exists to prevent accidental wipes.
    if not os.path.isdir(syspath(lib.directory)):
        ui.print_("Library path is unavailable or does not exist.")
        ui.print_(lib.directory)
        if not ui.input_yn("Are you sure you want to continue (y/n)?", True):
            return
    update_items(
        lib,
        args,
        opts.album,
        ui.should_move(opts.move),
        opts.pretend,
        opts.fields,
        opts.exclude_fields,
    )


update_cmd = ui.Subcommand(
    "update", help="update the library", aliases=("upd", "up")
)
update_cmd.parser.add_album_option()
update_cmd.parser.add_format_option()
update_cmd.parser.add_option(
    "-m",
    "--move",
    action="store_true",
    dest="move",
    help="move files in the library directory",
)
update_cmd.parser.add_option(
    "-M",
    "--nomove",
    action="store_false",
    dest="move",
    help="don't move files in library",
)
update_cmd.parser.add_option(
    "-p",
    "--pretend",
    action="store_true",
    help="show all changes but do nothing",
)
update_cmd.parser.add_option(
    "-F",
    "--field",
    default=None,
    action="append",
    dest="fields",
    help="list of fields to update",
)
update_cmd.parser.add_option(
    "-e",
    "--exclude-field",
    default=None,
    action="append",
    dest="exclude_fields",
    help="list of fields to exclude from updates",
)
update_cmd.func = update_func



"""Test the beets.export utilities associated with the export plugin."""

import json
import re  # used to test csv format
from xml.etree import ElementTree
from xml.etree.ElementTree import Element

from beets.test.helper import IOMixin, PluginTestCase


class ExportPluginTest(IOMixin, PluginTestCase):
    plugin = "export"

    def setUp(self):
        super().setUp()
        self.test_values = {"title": "xtitle", "album": "xalbum"}

    def execute_command(self, format_type, artist):
        query = ",".join(self.test_values.keys())
        return self.run_with_output(
            "export", "-f", format_type, "-i", query, artist
        )

    def create_item(self):
        (item,) = self.add_item_fixtures()
        item.artist = "xartist"
        item.title = self.test_values["title"]
        item.album = self.test_values["album"]
        item.write()
        item.store()
        return item

    def test_json_output(self):
        item1 = self.create_item()
        out = self.execute_command(format_type="json", artist=item1.artist)
        json_data = json.loads(out)[0]
        for key, val in self.test_values.items():
            assert key in json_data
            assert val == json_data[key]

    def test_jsonlines_output(self):
        item1 = self.create_item()
        out = self.execute_command(format_type="jsonlines", artist=item1.artist)
        json_data = json.loads(out)
        for key, val in self.test_values.items():
            assert key in json_data
            assert val == json_data[key]

    def test_csv_output(self):
        item1 = self.create_item()
        out = self.execute_command(format_type="csv", artist=item1.artist)
        csv_list = re.split("\r", re.sub("\n", "", out))
        head = re.split(",", csv_list[0])
        vals = re.split(",|\r", csv_list[1])
        for index, column in enumerate(head):
            assert self.test_values.get(column, None) is not None
            assert vals[index] == self.test_values[column]

    def test_xml_output(self):
        item1 = self.create_item()
        out = self.execute_command(format_type="xml", artist=item1.artist)
        library = ElementTree.fromstring(out)
        assert isinstance(library, Element)
        for track in library[0]:
            for details in track:
                tag = details.tag
                txt = details.text
                assert tag in self.test_values, tag
                assert self.test_values[tag] == txt, txt



import os
import platform
import time
from datetime import datetime

from beets.library import Item
from beets.test import _common
from beets.test.helper import IOMixin, PluginTestCase


def _parsetime(s):
    return time.mktime(datetime.strptime(s, "%Y-%m-%d %H:%M:%S").timetuple())


def _is_windows():
    return platform.system() == "Windows"


class MetaSyncTest(IOMixin, PluginTestCase):
    plugin = "metasync"
    itunes_library_unix = _common.RSRC / "itunes_library_unix.xml"
    itunes_library_windows = _common.RSRC / "itunes_library_windows.xml"

    def setUp(self):
        super().setUp()

        self.config["metasync"]["source"] = "itunes"

        if _is_windows():
            self.config["metasync"]["itunes"]["library"] = os.fsdecode(
                self.itunes_library_windows
            )
        else:
            self.config["metasync"]["itunes"]["library"] = os.fsdecode(
                self.itunes_library_unix
            )

        self._set_up_data()

    def _set_up_data(self):
        items = [_common.item() for _ in range(2)]

        items[0].title = "Tessellate"
        items[0].artist = "alt-J"
        items[0].albumartist = "alt-J"
        items[0].album = "An Awesome Wave"
        items[0].itunes_rating = 60

        items[1].title = "Breezeblocks"
        items[1].artist = "alt-J"
        items[1].albumartist = "alt-J"
        items[1].album = "An Awesome Wave"

        if _is_windows():
            items[
                0
            ].path = "G:\\Music\\Alt-J\\An Awesome Wave\\03 Tessellate.mp3"
            items[
                1
            ].path = "G:\\Music\\Alt-J\\An Awesome Wave\\04 Breezeblocks.mp3"
        else:
            items[0].path = "/Music/Alt-J/An Awesome Wave/03 Tessellate.mp3"
            items[1].path = "/Music/Alt-J/An Awesome Wave/04 Breezeblocks.mp3"

        for item in items:
            self.lib.add(item)

    def test_load_item_types(self):
        # This test also verifies that the MetaSources have loaded correctly
        assert "amarok_score" in Item._types
        assert "itunes_rating" in Item._types

    def test_pretend_sync_from_itunes(self):
        out = self.run_with_output("metasync", "-p")

        assert "itunes_rating: 60 -> 80" in out
        assert "itunes_rating: 100" in out
        assert "itunes_playcount: 31" in out
        assert "itunes_skipcount: 3" in out
        assert "itunes_lastplayed: 2015-05-04 12:20:51" in out
        assert "itunes_lastskipped: 2015-02-05 15:41:04" in out
        assert "itunes_dateadded: 2014-04-24 09:28:38" in out
        assert self.lib.items()[0].itunes_rating == 60

    def test_sync_from_itunes(self):
        self.run_command("metasync")

        assert self.lib.items()[0].itunes_rating == 80
        assert self.lib.items()[0].itunes_playcount == 0
        assert self.lib.items()[0].itunes_skipcount == 3
        assert not hasattr(self.lib.items()[0], "itunes_lastplayed")
        assert self.lib.items()[0].itunes_lastskipped == _parsetime(
            "2015-02-05 15:41:04"
        )
        assert self.lib.items()[0].itunes_dateadded == _parsetime(
            "2014-04-24 09:28:38"
        )

        assert self.lib.items()[1].itunes_rating == 100
        assert self.lib.items()[1].itunes_playcount == 31
        assert self.lib.items()[1].itunes_skipcount == 0
        assert self.lib.items()[1].itunes_lastplayed == _parsetime(
            "2015-05-04 12:20:51"
        )
        assert self.lib.items()[1].itunes_dateadded == _parsetime(
            "2014-04-24 09:28:38"
        )
        assert not hasattr(self.lib.items()[1], "itunes_lastskipped")



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from typing_extensions import NotRequired, TypedDict

if TYPE_CHECKING:
    from beets.autotag import TrackInfo


class ReleaseFormat(TypedDict):
    name: str
    qty: int
    descriptions: list[str] | None


class Artist(TypedDict):
    name: str
    anv: str
    join: str
    role: str
    tracks: str
    id: str
    resource_url: str


class AudioTrack(TypedDict):
    """Represent an audio item or logical segment in a Discogs tracklist.

    Discogs uses this shape both for independently addressable media tracks and
    for pieces rolled into one physical track. Position syntax and surrounding
    index structure determine which interpretation applies.
    """

    type_: Literal["track"]
    position: str
    title: str
    duration: str
    artists: NotRequired[list[Artist]]
    extraartists: NotRequired[list[Artist]]


class IndexTrack(TypedDict):
    """Represent a titled musical work containing related subtracks.

    The contained entries may be movements stored as separate physical tracks
    or logical divisions of one track. Their position syntax determines whether
    they remain separate or are coalesced under the work's metadata.
    """

    type_: Literal["index"]
    position: Literal[""]
    title: str
    duration: str
    sub_tracks: list[AudioTrack]
    artists: NotRequired[list[Artist]]
    extraartists: NotRequired[list[Artist]]


class HeadingTrack(TypedDict):
    """Represent descriptive tracklist structure rather than an audio item.

    A heading labels the following block but is not the title of a musical work.
    It must remain structural instead of being treated as an index container.
    """

    type_: Literal["heading"]
    position: Literal[""]
    title: str
    duration: str


Track = AudioTrack | IndexTrack | HeadingTrack


class ArtistInfo(TypedDict):
    artist: str
    artists: list[str]
    artist_credit: str
    artists_credit: list[str]
    artist_id: str
    artists_ids: list[str]
    arrangers: list[str] | None
    composers: list[str] | None
    remixers: list[str] | None
    lyricists: list[str] | None


class TracklistInfo(TypedDict):
    index: int
    index_tracks: dict[int, str]
    tracks: list[TrackInfo]
    divisions: list[str]
    next_divisions: list[str]
    mediums: list[str | None]
    medium_indices: list[str | None]



import time
from datetime import datetime

import pytest
from confuse import ConfigValueError

from beets.test.helper import IOMixin, PluginTestCase


class TypesPluginTest(IOMixin, PluginTestCase):
    plugin = "types"

    def test_integer_modify_and_query(self):
        self.config["types"] = {"myint": "int"}
        item = self.add_item(artist="aaa")

        # Do not match unset values
        out = self.list("myint:1..3")
        assert "" == out

        self.modify("myint=2")
        item.load()
        assert item["myint"] == 2

        # Match in range
        out = self.list("myint:1..3")
        assert "aaa" in out

    def test_album_integer_modify_and_query(self):
        self.config["types"] = {"myint": "int"}
        album = self.add_album(albumartist="aaa")

        # Do not match unset values
        out = self.list_album("myint:1..3")
        assert "" == out

        self.modify("-a", "myint=2")
        album.load()
        assert album["myint"] == 2

        # Match in range
        out = self.list_album("myint:1..3")
        assert "aaa" in out

    def test_float_modify_and_query(self):
        self.config["types"] = {"myfloat": "float"}
        item = self.add_item(artist="aaa")

        # Do not match unset values
        out = self.list("myfloat:10..0")
        assert "" == out

        self.modify("myfloat=-9.1")
        item.load()
        assert item["myfloat"] == -9.1

        # Match in range
        out = self.list("myfloat:-10..0")
        assert "aaa" in out

    def test_bool_modify_and_query(self):
        self.config["types"] = {"mybool": "bool"}
        true = self.add_item(artist="true")
        false = self.add_item(artist="false")
        self.add_item(artist="unset")

        # Do not match unset values
        out = self.list("mybool:true, mybool:false")
        assert "" == out

        # Set true
        self.modify("mybool=1", "artist:true")
        true.load()
        assert true["mybool"]

        # Set false
        self.modify("mybool=false", "artist:false")
        false.load()
        assert not false["mybool"]

        # Query bools
        out = self.list("mybool:true", "$artist $mybool")
        assert "true True" == out

        out = self.list("mybool:false", "$artist $mybool")

        # Dealing with unset fields?
        # assert 'false False' == out
        # out = self.list('mybool:', '$artist $mybool')
        # assert 'unset $mybool' in out

    def test_date_modify_and_query(self):
        self.config["types"] = {"mydate": "date"}
        # FIXME parsing should also work with default time format
        self.config["time_format"] = "%Y-%m-%d"
        old = self.add_item(artist="prince")
        new = self.add_item(artist="britney")

        # Do not match unset values
        out = self.list("mydate:..2000")
        assert "" == out

        self.modify("mydate=1999-01-01", "artist:prince")
        old.load()
        assert old["mydate"] == mktime(1999, 1, 1)

        self.modify("mydate=1999-12-30", "artist:britney")
        new.load()
        assert new["mydate"] == mktime(1999, 12, 30)

        # Match in range
        out = self.list("mydate:..1999-07", "$artist $mydate")
        assert "prince 1999-01-01" == out

        # FIXME some sort of timezone issue here
        # out = self.list('mydate:1999-12-30', '$artist $mydate')
        # assert 'britney 1999-12-30' == out

    def test_unknown_type_error(self):
        self.config["types"] = {"flex": "unkown type"}
        with pytest.raises(ConfigValueError):
            self.add_item(flex="test")

    def test_template_if_def(self):
        # Tests for a subtle bug when using %ifdef in templates along with
        # types that have truthy default values (e.g. '0', '0.0', 'False')
        # https://github.com/beetbox/beets/issues/3852
        self.config["types"] = {
            "playcount": "int",
            "rating": "float",
            "starred": "bool",
        }

        with_fields = self.add_item(artist="prince")
        self.modify("playcount=10", "artist=prince")
        self.modify("rating=5.0", "artist=prince")
        self.modify("starred=yes", "artist=prince")
        with_fields.load()

        without_fields = self.add_item(artist="britney")

        int_template = "%ifdef{playcount,Play count: $playcount,Not played}"
        assert with_fields.evaluate_template(int_template) == "Play count: 10"
        assert without_fields.evaluate_template(int_template) == "Not played"

        float_template = "%ifdef{rating,Rating: $rating,Not rated}"
        assert with_fields.evaluate_template(float_template) == "Rating: 5.0"
        assert without_fields.evaluate_template(float_template) == "Not rated"

        bool_template = "%ifdef{starred,Starred: $starred,Not starred}"
        assert with_fields.evaluate_template(bool_template).lower() in (
            "starred: true",
            "starred: yes",
            "starred: y",
        )
        assert without_fields.evaluate_template(bool_template) == "Not starred"

    def modify(self, *args):
        return self.run_with_output(
            "modify", "--yes", "--nowrite", "--nomove", *args
        )

    def list(self, query, fmt="$artist - $album - $title"):
        return self.run_with_output("ls", "-f", fmt, query).strip()

    def list_album(self, query, fmt="$albumartist - $album - $title"):
        return self.run_with_output("ls", "-a", "-f", fmt, query).strip()


def mktime(*args):
    return time.mktime(datetime(*args).timetuple())



from beets.util.pathformats import get_path_formats


def test_get_path_formats(config):
    """Ensure the function prepends custom and preserves defaults."""
    # override the default 'singleton' path and add a new one
    config["paths"].set({"singleton": "bar", "new": "hello"})

    path_formats = get_path_formats(config["paths"])
    actual_path_formats = [(key, tmpl.original) for key, tmpl in path_formats]
    assert actual_path_formats == [
        ("singleton:true", "bar"),
        ("new", "hello"),
        # defaults
        ("default", "$albumartist/$album%aunique{}/$track $title"),
        ("comp:true", "Compilations/$album%aunique{}/$track $title"),
    ]



"""Tests for the play plugin"""

import os
import sys
import unittest
from unittest.mock import ANY, patch

import pytest

from beets.test.helper import CleanupModulesMixin, IOMixin, PluginTestCase
from beets.ui import UserError
from beets.util import open_anything
from beetsplug.play import PlayPlugin


@patch("beetsplug.play.util.interactive_open")
class PlayPluginTest(IOMixin, CleanupModulesMixin, PluginTestCase):
    modules = (PlayPlugin.__module__,)
    plugin = "play"

    def setUp(self):
        super().setUp()
        self.item = self.add_item(album="a nice älbum", title="aNiceTitle")
        self.lib.add_album([self.item])
        self.config["play"]["command"] = "echo"

    def run_and_assert(
        self,
        open_mock,
        args=("title:aNiceTitle",),
        expected_cmd="echo",
        expected_playlist=None,
    ):
        self.run_command("play", *args)

        open_mock.assert_called_once_with(ANY, expected_cmd)
        expected_playlist = expected_playlist or self.item.path.decode("utf-8")
        exp_playlist = f"{expected_playlist}\n"
        with open(open_mock.call_args[0][0][0], "rb") as playlist:
            assert exp_playlist == playlist.read().decode("utf-8")

    def test_basic(self, open_mock):
        self.run_and_assert(open_mock)

    def test_album_option(self, open_mock):
        self.run_and_assert(open_mock, ["-a", "nice"])

    def test_args_option(self, open_mock):
        self.run_and_assert(
            open_mock, ["-A", "foo", "title:aNiceTitle"], "echo foo"
        )

    def test_args_option_in_middle(self, open_mock):
        self.config["play"]["command"] = "echo $args other"

        self.run_and_assert(
            open_mock, ["-A", "foo", "title:aNiceTitle"], "echo foo other"
        )

    def test_unset_args_option_in_middle(self, open_mock):
        self.config["play"]["command"] = "echo $args other"

        self.run_and_assert(open_mock, ["title:aNiceTitle"], "echo other")

    # FIXME: fails on windows
    @unittest.skipIf(sys.platform == "win32", "win32")
    def test_relative_to(self, open_mock):
        self.config["play"]["command"] = "echo"
        self.config["play"]["relative_to"] = "/something"

        path = os.path.relpath(self.item.path, b"/something")
        playlist = path.decode("utf-8")
        self.run_and_assert(
            open_mock, expected_cmd="echo", expected_playlist=playlist
        )

    def test_use_folders(self, open_mock):
        self.config["play"]["command"] = None
        self.config["play"]["use_folders"] = True
        self.run_command("play", "-a", "nice")

        open_mock.assert_called_once_with(ANY, open_anything())
        with open(open_mock.call_args[0][0][0], "rb") as f:
            playlist = f.read().decode("utf-8")
        assert f"{self.item.filepath.parent}\n" == playlist

    def test_raw(self, open_mock):
        self.config["play"]["raw"] = True

        self.run_command("play", "nice")

        open_mock.assert_called_once_with([self.item.path], "echo")

    def test_pls_marker(self, open_mock):
        self.config["play"]["command"] = (
            "echo --some params --playlist=$playlist --some-more params"
        )

        self.run_command("play", "nice")

        open_mock.assert_called_once

        commandstr = open_mock.call_args_list[0][0][1]
        assert commandstr.startswith("echo --some params --playlist=")
        assert commandstr.endswith(" --some-more params")

    def test_not_found(self, open_mock):
        self.run_command("play", "not found")

        open_mock.assert_not_called()

    def test_warning_threshold(self, open_mock):
        self.config["play"]["warning_threshold"] = 1
        self.add_item(title="another NiceTitle")

        self.io.addinput("a")
        self.run_command("play", "nice")

        open_mock.assert_not_called()

    def test_skip_warning_threshold_bypass(self, open_mock):
        self.config["play"]["warning_threshold"] = 1
        self.other_item = self.add_item(title="another NiceTitle")

        expected_playlist = f"{self.item.filepath}\n{self.other_item.filepath}"

        self.io.addinput("a")
        self.run_and_assert(
            open_mock, ["-y", "NiceTitle"], expected_playlist=expected_playlist
        )

    def _playlist_lines(self, open_mock):
        """Read the playlist file passed to interactive_open and return its lines."""
        # interactive_open is called as: interactive_open([playlist_path], command)
        playlist_path = open_mock.call_args[0][0][0]
        with open(playlist_path, "rb") as playlist:
            return playlist.read().decode("utf-8").splitlines()

    def _add_many_ordered_items(self, *, count, album):
        items = []
        for track in range(1, count + 1):
            items.append(
                self.add_item(
                    album=album,
                    artist="randomize artist",
                    title=f"randomize {track:03d}",
                    track=track,
                )
            )
        return items

    def test_randomize(self, open_mock):
        album = "randomize_test"
        items = self._add_many_ordered_items(count=100, album=album)
        baseline = [str(item.filepath) for item in items]

        self.run_command("play", "-R", f"album:{album}")
        lines = self._playlist_lines(open_mock)
        assert sorted(lines) == sorted(baseline), (
            "playlist items are not the same"
        )
        assert lines != baseline, "playlist order hasn't changed"

    def test_command_failed(self, open_mock):
        open_mock.side_effect = OSError("some reason")

        with pytest.raises(UserError):
            self.run_command("play", "title:aNiceTitle")



"""Shows file metadata."""

import os

import mediafile

from beets import ui
from beets.library import Item
from beets.plugins import BeetsPlugin
from beets.util import displayable_path, normpath, syspath


def tag_data(lib, args, album=False):
    query = []
    for arg in args:
        path = normpath(arg)
        if os.path.isfile(syspath(path)):
            yield tag_data_emitter(path)
        else:
            query.append(arg)

    if query:
        for item in lib.items(query):
            yield tag_data_emitter(item.path)


def tag_fields():
    fields = set(mediafile.MediaFile.readable_fields())
    fields.add("art")
    return fields


def tag_data_emitter(path):
    def emitter(included_keys):
        if included_keys == "*":
            fields = tag_fields()
        else:
            fields = included_keys
        if "images" in fields:
            # We can't serialize the image data.
            fields.remove("images")
        mf = mediafile.MediaFile(syspath(path))
        tags = {}
        for field in fields:
            if field == "art":
                tags[field] = mf.art is not None
            else:
                tags[field] = getattr(mf, field, None)

        # create a temporary Item to take advantage of __format__
        item = Item.from_path(syspath(path))

        return tags, item

    return emitter


def library_data(lib, args, album=False):
    for item in lib.albums(args) if album else lib.items(args):
        yield library_data_emitter(item)


def library_data_emitter(item):
    def emitter(included_keys):
        data = dict(item.formatted(included_keys=included_keys))

        return data, item

    return emitter


def update_summary(summary, tags):
    for key, value in tags.items():
        if key not in summary:
            summary[key] = value
        elif summary[key] != value:
            summary[key] = "[various]"
    return summary


def print_data(data, item=None, fmt=None):
    """Print, with optional formatting, the fields of a single element.

    If no format string `fmt` is passed, the entries on `data` are printed one
    in each line, with the format 'field: value'. If `fmt` is not `None`, the
    `item` is printed according to `fmt`, using the `Item.__format__`
    machinery.
    """
    if fmt:
        # use fmt specified by the user
        ui.print_(format(item, fmt))
        return

    path = displayable_path(item.path) if item else None
    formatted = {}
    for key, value in data.items():
        if isinstance(value, list):
            formatted[key] = "; ".join(value)
        if value is not None:
            formatted[key] = value

    if len(formatted) == 0:
        return

    maxwidth = max(len(key) for key in formatted)

    if path:
        ui.print_(displayable_path(path))

    for field in sorted(formatted):
        value = formatted[field]
        if isinstance(value, list):
            value = "; ".join(value)
        ui.print_(f"{field:>{maxwidth}}: {value}")


def print_data_keys(data, item=None):
    """Print only the keys (field names) for an item."""
    path = displayable_path(item.path) if item else None
    formatted = []
    for key, value in data.items():
        formatted.append(key)

    if len(formatted) == 0:
        return

    if path:
        ui.print_(displayable_path(path))

    for field in sorted(formatted):
        ui.print_(f"    {field}")


class InfoPlugin(BeetsPlugin):
    def commands(self):
        cmd = ui.Subcommand("info", help="show file metadata")
        cmd.func = self.run
        cmd.parser.add_option(
            "-l",
            "--library",
            action="store_true",
            help="show library fields instead of tags",
        )
        cmd.parser.add_option(
            "-a",
            "--album",
            action="store_true",
            help='show album fields instead of tracks (implies "--library")',
        )
        cmd.parser.add_option(
            "-s",
            "--summarize",
            action="store_true",
            help="summarize the tags of all files",
        )
        cmd.parser.add_option(
            "-i",
            "--include-keys",
            default=[],
            action="append",
            dest="included_keys",
            help="comma separated list of keys to show",
        )
        cmd.parser.add_option(
            "-k", "--keys-only", action="store_true", help="show only the keys"
        )
        cmd.parser.add_format_option(target="item")
        return [cmd]

    def run(self, lib, opts, args):
        """Print tag info or library data for each file referenced by args.

        Main entry point for the `beet info ARGS...` command.

        If an argument is a path pointing to an existing file, then the tags
        of that file are printed. All other arguments are considered
        queries, and for each item matching all those queries the tags from
        the file are printed.

        If `opts.summarize` is true, the function merges all tags into one
        dictionary and only prints that. If two files have different values
        for the same tag, the value is set to '[various]'
        """
        if opts.library or opts.album:
            data_collector = library_data
        else:
            data_collector = tag_data

        included_keys = []
        for keys in opts.included_keys:
            included_keys.extend(keys.split(","))
        # Drop path even if user provides it multiple times
        included_keys = [k for k in included_keys if k != "path"]

        first = True
        summary = {}
        for data_emitter in data_collector(lib, args, album=opts.album):
            try:
                data, item = data_emitter(included_keys or "*")
            except (mediafile.UnreadableFileError, OSError) as ex:
                self._log.error("cannot read file: {}", ex)
                continue

            if opts.summarize:
                update_summary(summary, data)
            else:
                if not first:
                    ui.print_()
                if opts.keys_only:
                    print_data_keys(data, item)
                else:
                    fmt = [opts.format][0] if opts.format else None
                    print_data(data, item, fmt)
                first = False

        if opts.summarize:
            print_data(summary)



"""The `import` command: import new music into the library."""

import os

from beets import config, logging, plugins, ui
from beets.exceptions import UserError
from beets.util import displayable_path, normpath, syspath

from .session import TerminalImportSession

# Global logger.
log = logging.getLogger("beets")


def paths_from_logfile(path):
    """Parse the logfile and yield skipped paths to pass to the `import`
    command.
    """
    with open(path, encoding="utf-8") as fp:
        for i, line in enumerate(fp, start=1):
            verb, sep, paths = line.rstrip("\n").partition(" ")
            if not sep:
                raise ValueError(f"line {i} is invalid")

            # Ignore informational lines that don't need to be re-imported.
            if verb in {"import", "duplicate-keep", "duplicate-replace"}:
                continue

            if verb not in {"asis", "skip", "duplicate-skip"}:
                raise ValueError(f"line {i} contains unknown verb {verb}")

            yield os.path.commonpath(paths.split("; "))


def parse_logfiles(logfiles):
    """Parse all `logfiles` and yield paths from it."""
    for logfile in logfiles:
        try:
            yield from paths_from_logfile(syspath(normpath(logfile)))
        except ValueError as err:
            raise UserError(
                f"malformed logfile {displayable_path(logfile)}: {err}"
            ) from err
        except OSError as err:
            raise UserError(
                f"unreadable logfile {displayable_path(logfile)}: {err}"
            ) from err


def import_files(lib, paths: list[bytes], query):
    """Import the files in the given list of paths or matching the
    query.
    """
    # Check parameter consistency.
    if config["import"]["quiet"] and config["import"]["timid"]:
        raise UserError("can't be both quiet and timid")

    # Open the log.
    if config["import"]["log"].get() is not None:
        logpath = syspath(config["import"]["log"].as_filename())
        try:
            loghandler = logging.FileHandler(logpath, encoding="utf-8")
        except OSError:
            raise UserError(
                "Could not open log file for writing:"
                f" {displayable_path(logpath)}"
            )
    else:
        loghandler = None

    # Never ask for input in quiet mode.
    if config["import"]["resume"].get() == "ask" and config["import"]["quiet"]:
        config["import"]["resume"] = False

    session = TerminalImportSession(lib, loghandler, paths, query)
    session.run()

    # Emit event.
    plugins.send("import", lib=lib, paths=paths)


def import_func(lib, opts, args: list[str]):
    config["import"].set_args(opts)

    # Special case: --copy flag suppresses import_move (which would
    # otherwise take precedence).
    if opts.copy:
        config["import"]["move"] = False

    if opts.library:
        query = args
        byte_paths = []
    else:
        query = None
        paths = args

        # The paths from the logfiles go into a separate list to allow handling
        # errors differently from user-specified paths.
        paths_from_logfiles = list(parse_logfiles(opts.from_logfiles or []))

        if not paths and not paths_from_logfiles:
            raise UserError("no path specified")

        byte_paths = [os.fsencode(p) for p in paths]
        paths_from_logfiles = [os.fsencode(p) for p in paths_from_logfiles]

        # Check the user-specified directories.
        for path in byte_paths:
            if not os.path.exists(syspath(normpath(path))):
                raise UserError(
                    f"no such file or directory: {displayable_path(path)}"
                )

        # Check the directories from the logfiles, but don't throw an error in
        # case those paths don't exist. Maybe some of those paths have already
        # been imported and moved separately, so logging a warning should
        # suffice.
        for path in paths_from_logfiles:
            if not os.path.exists(syspath(normpath(path))):
                log.warning(
                    "No such file or directory: {}", displayable_path(path)
                )
                continue

            byte_paths.append(path)

        # If all paths were read from a logfile, and none of them exist, throw
        # an error
        if not byte_paths:
            raise UserError("none of the paths are importable")

    import_files(lib, byte_paths, query)


def _store_dict(option, opt_str, value, parser):
    """Custom action callback to parse options which have ``key=value``
    pairs as values. All such pairs passed for this option are
    aggregated into a dictionary.
    """
    dest = option.dest
    option_values = getattr(parser.values, dest, None)

    if option_values is None:
        # This is the first supplied ``key=value`` pair of option.
        # Initialize empty dictionary and get a reference to it.
        setattr(parser.values, dest, {})
        option_values = getattr(parser.values, dest)

    try:
        key, value = value.split("=", 1)
        if not (key and value):
            raise ValueError
    except ValueError:
        raise UserError(
            f"supplied argument `{value}' is not of the form `key=value'"
        )

    option_values[key] = value


import_cmd = ui.Subcommand(
    "import", help="import new music", aliases=("imp", "im")
)
import_cmd.parser.add_option(
    "-c",
    "--copy",
    action="store_true",
    default=None,
    help="copy tracks into library directory (default)",
)
import_cmd.parser.add_option(
    "-C",
    "--nocopy",
    action="store_false",
    dest="copy",
    help="don't copy tracks (opposite of -c)",
)
import_cmd.parser.add_option(
    "-m",
    "--move",
    action="store_true",
    dest="move",
    help="move tracks into the library (overrides -c)",
)
import_cmd.parser.add_option(
    "-M",
    "--nomove",
    action="store_false",
    dest="move",
    help="don't move tracks into the library (overrides -m)",
)
import_cmd.parser.add_option(
    "-w",
    "--write",
    action="store_true",
    default=None,
    help="write new metadata to files' tags (default)",
)
import_cmd.parser.add_option(
    "-W",
    "--nowrite",
    action="store_false",
    dest="write",
    help="don't write metadata (opposite of -w)",
)
import_cmd.parser.add_option(
    "-a",
    "--autotag",
    action="store_true",
    dest="autotag",
    help="infer tags for imported files (default)",
)
import_cmd.parser.add_option(
    "-A",
    "--noautotag",
    action="store_false",
    dest="autotag",
    help="don't infer tags for imported files (opposite of -a)",
)
import_cmd.parser.add_option(
    "-p",
    "--resume",
    action="store_true",
    default=None,
    help="resume importing if interrupted",
)
import_cmd.parser.add_option(
    "-P",
    "--noresume",
    action="store_false",
    dest="resume",
    help="do not try to resume importing",
)
import_cmd.parser.add_option(
    "-q",
    "--quiet",
    action="store_true",
    dest="quiet",
    help="never prompt for input: skip albums instead",
)
import_cmd.parser.add_option(
    "--quiet-fallback",
    type="string",
    dest="quiet_fallback",
    help="decision in quiet mode when no strong match: skip or asis",
)
import_cmd.parser.add_option(
    "-l",
    "--log",
    dest="log",
    help="file to log untaggable albums for later review",
)
import_cmd.parser.add_option(
    "-s",
    "--singletons",
    action="store_true",
    help="import individual tracks instead of full albums",
)
import_cmd.parser.add_option(
    "-t",
    "--timid",
    dest="timid",
    action="store_true",
    help="always confirm all actions",
)
import_cmd.parser.add_option(
    "-L",
    "--library",
    dest="library",
    action="store_true",
    help="retag items matching a query",
)
import_cmd.parser.add_option(
    "-i",
    "--incremental",
    dest="incremental",
    action="store_true",
    help="skip already-imported directories",
)
import_cmd.parser.add_option(
    "-I",
    "--noincremental",
    dest="incremental",
    action="store_false",
    help="do not skip already-imported directories",
)
import_cmd.parser.add_option(
    "-R",
    "--incremental-skip-later",
    action="store_true",
    dest="incremental_skip_later",
    help="do not record skipped files during incremental import",
)
import_cmd.parser.add_option(
    "-r",
    "--noincremental-skip-later",
    action="store_false",
    dest="incremental_skip_later",
    help="record skipped files during incremental import",
)
import_cmd.parser.add_option(
    "--from-scratch",
    dest="from_scratch",
    action="store_true",
    help="erase existing metadata before applying new metadata",
)
import_cmd.parser.add_option(
    "--flat",
    dest="flat",
    action="store_true",
    help="import an entire tree as a single album",
)
import_cmd.parser.add_option(
    "-g",
    "--group-albums",
    dest="group_albums",
    action="store_true",
    help="group tracks in a folder into separate albums",
)
import_cmd.parser.add_option(
    "--pretend",
    dest="pretend",
    action="store_true",
    help="just print the files to import",
)
import_cmd.parser.add_option(
    "-S",
    "--search-id",
    dest="search_ids",
    action="append",
    metavar="ID",
    help="restrict matching to a specific metadata backend ID",
)
import_cmd.parser.add_option(
    "--from-logfile",
    dest="from_logfiles",
    action="append",
    metavar="PATH",
    help="read skipped paths from an existing logfile",
)
import_cmd.parser.add_option(
    "--set",
    dest="set_fields",
    action="callback",
    callback=_store_dict,
    metavar="FIELD=VALUE",
    help="set the given fields to the supplied values",
)
import_cmd.func = import_func



import os.path
from shutil import rmtree
from tempfile import mkdtemp
from unittest.mock import Mock, call, patch

import pytest

from beets.test.helper import BeetsTestCase
from beets.util import bytestring_path, syspath
from beetsplug.thumbnails import (
    LARGE_DIR,
    NORMAL_DIR,
    GioURI,
    PathlibURI,
    ThumbnailsPlugin,
)


class ThumbnailsTest(BeetsTestCase):
    @patch("beetsplug.thumbnails.ArtResizer")
    @patch("beetsplug.thumbnails.ThumbnailsPlugin._check_local_ok", Mock())
    @patch("beetsplug.thumbnails.os.stat")
    def test_add_tags(self, mock_stat, mock_artresizer):
        plugin = ThumbnailsPlugin()
        plugin.get_uri = Mock(
            side_effect={b"/path/to/cover": "COVER_URI"}.__getitem__
        )
        album = Mock(artpath=b"/path/to/cover")
        mock_stat.return_value.st_mtime = 12345

        plugin.add_tags(album, b"/path/to/thumbnail")

        metadata = {"Thumb::URI": "COVER_URI", "Thumb::MTime": "12345"}
        mock_artresizer.shared.write_metadata.assert_called_once_with(
            b"/path/to/thumbnail", metadata
        )
        mock_stat.assert_called_once_with(syspath(album.artpath))

    @patch("beetsplug.thumbnails.os")
    @patch("beetsplug.thumbnails.ArtResizer")
    @patch("beetsplug.thumbnails.GioURI")
    def test_check_local_ok(self, mock_giouri, mock_artresizer, mock_os):
        # test local resizing capability
        mock_artresizer.shared.local = False
        mock_artresizer.shared.can_write_metadata = False
        plugin = ThumbnailsPlugin()
        assert not plugin._check_local_ok()

        # test dirs creation
        mock_artresizer.shared.local = True
        mock_artresizer.shared.can_write_metadata = True

        def exists(path):
            if path == syspath(NORMAL_DIR):
                return False
            if path == syspath(LARGE_DIR):
                return True
            raise ValueError(f"unexpected path {path!r}")

        mock_os.path.exists = exists
        plugin = ThumbnailsPlugin()
        mock_os.makedirs.assert_called_once_with(syspath(NORMAL_DIR))
        assert plugin._check_local_ok()

        # test metadata writer function
        mock_os.path.exists = lambda _: True

        mock_artresizer.shared.local = True
        mock_artresizer.shared.can_write_metadata = False
        with pytest.raises(RuntimeError):
            ThumbnailsPlugin()

        mock_artresizer.shared.local = True
        mock_artresizer.shared.can_write_metadata = True
        assert ThumbnailsPlugin()._check_local_ok()

        # test URI getter function
        giouri_inst = mock_giouri.return_value
        giouri_inst.available = True
        assert ThumbnailsPlugin().get_uri == giouri_inst.uri

        giouri_inst.available = False
        assert ThumbnailsPlugin().get_uri.__self__.__class__ == PathlibURI

    @patch("beetsplug.thumbnails.ThumbnailsPlugin._check_local_ok", Mock())
    @patch("beetsplug.thumbnails.ArtResizer")
    @patch("beets.util.syspath", Mock(side_effect=lambda x: x))
    @patch("beetsplug.thumbnails.os")
    @patch("beetsplug.thumbnails.shutil")
    def test_make_cover_thumbnail(self, mock_shutils, mock_os, mock_artresizer):
        thumbnail_dir = os.path.normpath(b"/thumbnail/dir")
        md5_file = os.path.join(thumbnail_dir, b"md5")
        path_to_art = os.path.normpath(b"/path/to/art")
        path_to_resized_art = os.path.normpath(b"/path/to/resized/artwork")

        mock_os.path.join = os.path.join  # don't mock that function
        plugin = ThumbnailsPlugin()
        plugin.add_tags = Mock()

        album = Mock(artpath=path_to_art)
        plugin.thumbnail_file_name = Mock(return_value=b"md5")
        mock_os.path.exists.return_value = False

        mock_resize = mock_artresizer.shared.resize
        mock_resize.return_value = path_to_resized_art

        plugin.make_cover_thumbnail(album, 12345, thumbnail_dir)

        mock_os.path.exists.assert_called_once_with(syspath(md5_file))

        mock_resize.assert_called_once_with(12345, path_to_art, md5_file)
        plugin.add_tags.assert_called_once_with(album, path_to_resized_art)
        mock_shutils.move.assert_called_once_with(
            syspath(path_to_resized_art), syspath(md5_file)
        )

        # now test with recent thumbnail & with force
        mock_os.path.exists.return_value = True
        plugin.force = False
        mock_resize.reset_mock()

        def os_stat(target):
            if target == syspath(md5_file):
                return Mock(st_mtime=3)
            if target == syspath(path_to_art):
                return Mock(st_mtime=2)
            raise ValueError(f"invalid target {target}")

        mock_os.stat.side_effect = os_stat

        plugin.make_cover_thumbnail(album, 12345, thumbnail_dir)
        assert mock_resize.call_count == 0

        # and with force
        plugin.config["force"] = True
        plugin.make_cover_thumbnail(album, 12345, thumbnail_dir)
        mock_resize.assert_called_once_with(12345, path_to_art, md5_file)

    @patch("beetsplug.thumbnails.ThumbnailsPlugin._check_local_ok", Mock())
    def test_make_dolphin_cover_thumbnail(self):
        plugin = ThumbnailsPlugin()
        tmp = bytestring_path(mkdtemp())
        album = Mock(path=tmp, artpath=os.path.join(tmp, b"cover.jpg"))
        plugin.make_dolphin_cover_thumbnail(album)
        with open(os.path.join(tmp, b".directory"), "rb") as f:
            assert f.read().splitlines() == [
                b"[Desktop Entry]",
                b"Icon=./cover.jpg",
            ]

        # not rewritten when it already exists (yup that's a big limitation)
        album.artpath = b"/my/awesome/art.tiff"
        plugin.make_dolphin_cover_thumbnail(album)
        with open(os.path.join(tmp, b".directory"), "rb") as f:
            assert f.read().splitlines() == [
                b"[Desktop Entry]",
                b"Icon=./cover.jpg",
            ]

        rmtree(syspath(tmp))

    @patch("beetsplug.thumbnails.ThumbnailsPlugin._check_local_ok", Mock())
    @patch("beetsplug.thumbnails.ArtResizer")
    def test_process_album(self, mock_artresizer):
        get_size = mock_artresizer.shared.get_size

        plugin = ThumbnailsPlugin()
        make_cover = plugin.make_cover_thumbnail = Mock(return_value=True)
        make_dolphin = plugin.make_dolphin_cover_thumbnail = Mock()

        # no art
        album = Mock(artpath=None)
        plugin.process_album(album)
        assert get_size.call_count == 0
        assert make_dolphin.call_count == 0

        # cannot get art size
        album.artpath = b"/path/to/art"
        get_size.return_value = None
        plugin.process_album(album)
        get_size.assert_called_once_with(b"/path/to/art")
        assert make_cover.call_count == 0

        # dolphin tests
        plugin.config["dolphin"] = False
        plugin.process_album(album)
        assert make_dolphin.call_count == 0

        plugin.config["dolphin"] = True
        plugin.process_album(album)
        make_dolphin.assert_called_once_with(album)

        # small art
        get_size.return_value = 200, 200
        plugin.process_album(album)
        make_cover.assert_called_once_with(album, 128, NORMAL_DIR)

        # big art
        make_cover.reset_mock()
        get_size.return_value = 500, 500
        plugin.process_album(album)
        make_cover.assert_has_calls(
            [call(album, 128, NORMAL_DIR), call(album, 256, LARGE_DIR)],
            any_order=True,
        )

    @patch("beetsplug.thumbnails.ThumbnailsPlugin._check_local_ok", Mock())
    def test_invokations(self):
        plugin = ThumbnailsPlugin()
        plugin.process_album = Mock()
        album = Mock()

        plugin.process_album.reset_mock()
        lib = Mock()
        album2 = Mock()
        lib.albums.return_value = [album, album2]
        plugin.process_query(lib, Mock(), None)
        plugin.process_album.assert_has_calls(
            [call(album), call(album2)], any_order=True
        )

    @patch("beetsplug.thumbnails.BaseDirectory")
    def test_thumbnail_file_name(self, mock_basedir):
        plug = ThumbnailsPlugin()
        plug.get_uri = Mock(return_value="file:///my/uri")
        assert (
            plug.thumbnail_file_name(b"idontcare")
            == b"9488f5797fbe12ffb316d607dfd93d04.png"
        )

    def test_uri(self):
        gio = GioURI()
        if not gio.available:
            self.skipTest("GIO library not found")

        import ctypes

        with pytest.raises(ctypes.ArgumentError):
            gio.uri("/foo")
        assert gio.uri(b"/foo") == "file:///foo"
        assert gio.uri(b"/foo!") == "file:///foo!"
        assert (
            gio.uri(b"/music/\xec\x8b\xb8\xec\x9d\xb4")
            == "file:///music/%EC%8B%B8%EC%9D%B4"
        )


class TestPathlibURI:
    """Test PathlibURI class"""

    def test_uri(self):
        test_uri = PathlibURI()

        # test it won't break if we pass it bytes for a path
        test_uri.uri(b"/")



"""List duplicate tracks or albums."""

import os
import shlex

from beets.library import Album, Item
from beets.plugins import BeetsPlugin
from beets.ui import Subcommand, UserError, print_
from beets.util import (
    MoveOperation,
    bytestring_path,
    command_output,
    displayable_path,
    subprocess,
)

PLUGIN = "duplicates"


class DuplicatesPlugin(BeetsPlugin):
    """List duplicate tracks or albums"""

    def __init__(self):
        super().__init__()

        self.config.add(
            {
                "album": False,
                "checksum": "",
                "copy": "",
                "count": False,
                "delete": False,
                "format": "",
                "full": False,
                "keys": [],
                "merge": False,
                "move": "",
                "path": False,
                "tiebreak": {},
                "strict": False,
                "tag": "",
                "remove": False,
            }
        )

        self._command = Subcommand("duplicates", help=__doc__, aliases=["dup"])
        self._command.parser.add_option(
            "-c",
            "--count",
            dest="count",
            action="store_true",
            help="show duplicate counts",
        )
        self._command.parser.add_option(
            "-C",
            "--checksum",
            dest="checksum",
            action="store",
            metavar="PROG",
            help="report duplicates based on arbitrary command",
        )
        self._command.parser.add_option(
            "-d",
            "--delete",
            dest="delete",
            action="store_true",
            help="delete items from library and disk",
        )
        self._command.parser.add_option(
            "-F",
            "--full",
            dest="full",
            action="store_true",
            help="show all versions of duplicate tracks or albums",
        )
        self._command.parser.add_option(
            "-s",
            "--strict",
            dest="strict",
            action="store_true",
            help="report duplicates only if all attributes are set",
        )
        self._command.parser.add_option(
            "-k",
            "--key",
            dest="keys",
            action="append",
            metavar="KEY",
            help="report duplicates based on keys (use multiple times)",
        )
        self._command.parser.add_option(
            "-M",
            "--merge",
            dest="merge",
            action="store_true",
            help="merge duplicate items",
        )
        self._command.parser.add_option(
            "-m",
            "--move",
            dest="move",
            action="store",
            metavar="DEST",
            help="move items to dest",
        )
        self._command.parser.add_option(
            "-o",
            "--copy",
            dest="copy",
            action="store",
            metavar="DEST",
            help="copy items to dest",
        )
        self._command.parser.add_option(
            "-t",
            "--tag",
            dest="tag",
            action="store",
            help="tag matched items with 'k=v' attribute",
        )
        self._command.parser.add_option(
            "-r",
            "--remove",
            dest="remove",
            action="store_true",
            help="remove items from library",
        )
        self._command.parser.add_all_common_options()

    def commands(self):
        def _dup(lib, opts, args):
            self.config.set_args(opts)
            album = self.config["album"].get(bool)
            checksum = self.config["checksum"].get(str)
            copy = bytestring_path(self.config["copy"].as_str())
            count = self.config["count"].get(bool)
            delete = self.config["delete"].get(bool)
            remove = self.config["remove"].get(bool)
            fmt_tmpl = self.config["format"].get(str)
            full = self.config["full"].get(bool)
            keys = self.config["keys"].as_str_seq()
            merge = self.config["merge"].get(bool)
            move = bytestring_path(self.config["move"].as_str())
            path = self.config["path"].get(bool)
            tiebreak = self.config["tiebreak"].get(dict)
            strict = self.config["strict"].get(bool)
            tag = self.config["tag"].get(str)

            if album:
                if not keys:
                    keys = ["mb_albumid"]
                items = lib.albums(args)
            else:
                if not keys:
                    keys = ["mb_trackid", "mb_albumid"]
                items = lib.items(args)

            # If there's nothing to do, return early. The code below assumes
            # `items` to be non-empty.
            if not items:
                return

            if path:
                fmt_tmpl = "$path"
            elif not fmt_tmpl:
                if album:
                    fmt_tmpl = "$albumartist - $album"
                else:
                    fmt_tmpl = "$albumartist - $album - $title"

            if checksum:
                for i in items:
                    k, _ = self._checksum(i, checksum)
                keys = [k]

            for obj_id, obj_count, objs in self._duplicates(
                items,
                keys=keys,
                full=full,
                strict=strict,
                tiebreak=tiebreak,
                merge=merge,
            ):
                if obj_id:  # Skip empty IDs.
                    for o in objs:
                        self._process_item(
                            o,
                            copy=copy,
                            move=move,
                            delete=delete,
                            remove=remove,
                            tag=tag,
                            fmt=(
                                fmt_tmpl
                                if not count
                                else f"{fmt_tmpl}: {obj_count}"
                            ),
                        )

        self._command.func = _dup
        return [self._command]

    def _process_item(
        self,
        item,
        copy=False,
        move=False,
        delete=False,
        tag=False,
        fmt="",
        remove=False,
    ):
        """Process Item `item`."""
        print_(format(item, fmt))
        if copy:
            item.move(basedir=copy, operation=MoveOperation.COPY)
            item.store()
        if move:
            item.move(basedir=move)
            item.store()
        if delete:
            item.remove(delete=True)
        elif remove:
            item.remove(delete=False)
        if tag:
            try:
                k, v = tag.split("=")
            except Exception:
                raise UserError(f"{PLUGIN}: can't parse k=v tag: {tag}")
            setattr(item, k, v)
            item.store()

    def _checksum(self, item, prog):
        """Run external `prog` on file path associated with `item`, cache
        output as flexattr on a key that is the name of the program, and
        return the key, checksum tuple.
        """
        args = [
            p.format(file=os.fsdecode(item.path)) for p in shlex.split(prog)
        ]
        key = args[0]
        checksum = getattr(item, key, False)
        if not checksum:
            self._log.debug(
                "key {} on item {.filepath} not cached:computing checksum",
                key,
                item,
            )
            try:
                checksum = command_output(args).stdout
                setattr(item, key, checksum)
                item.store()
                self._log.debug(
                    "computed checksum for {.title} using {}", item, key
                )
            except subprocess.CalledProcessError as e:
                self._log.debug("failed to checksum {.filepath}: {}", item, e)
        else:
            self._log.debug(
                "key {} on item {.filepath} cached:not computing checksum",
                key,
                item,
            )
        return key, checksum

    def _group_by(self, objs, keys, strict):
        """Return a dictionary with keys arbitrary concatenations of attributes
        and values lists of objects (Albums or Items) with those keys.

        If strict, all attributes must be defined for a duplicate match.
        """
        import collections

        counts = collections.defaultdict(list)
        for obj in objs:
            values = [getattr(obj, k, None) for k in keys]
            values = [v for v in values if v not in (None, "")]
            if strict and len(values) < len(keys):
                self._log.debug(
                    "some keys {} on item {.filepath} are null or empty: skipping",
                    keys,
                    obj,
                )
            elif not strict and not len(values):
                self._log.debug(
                    "all keys {} on item {.filepath} are null or empty: skipping",
                    keys,
                    obj,
                )
            else:
                key = tuple(values)
                counts[key].append(obj)

        return counts

    def _order(self, objs, tiebreak=None):
        """Return the objects (Items or Albums) sorted by descending
        order of priority.

        If provided, the `tiebreak` dict indicates the field to use to
        prioritize the objects. Otherwise, Items are placed in order of
        "completeness" (objects with more non-null fields come first)
        and Albums are ordered by their track count.
        """
        kind = "items" if all(isinstance(o, Item) for o in objs) else "albums"

        if tiebreak and kind in tiebreak.keys():

            def key(x):
                return tuple(getattr(x, k) for k in tiebreak[kind])
        else:
            if kind == "items":

                def truthy(v):
                    # Avoid a Unicode warning by avoiding comparison
                    # between a bytes object and the empty Unicode
                    # string ''.
                    return v is not None and (
                        v != "" if isinstance(v, str) else True
                    )

                fields = Item.all_keys()

                def key(x):
                    return sum(1 for f in fields if truthy(getattr(x, f)))
            else:

                def key(x):
                    return len(x.items())

        return sorted(objs, key=key, reverse=True)

    def _merge_items(self, objs):
        """Merge Item objs by copying missing fields from items in the tail to
        the head item.

        Return same number of items, with the head item modified.
        """
        fields = Item.all_keys()
        for f in fields:
            for o in objs[1:]:
                if getattr(objs[0], f, None) in (None, ""):
                    value = getattr(o, f, None)
                    if value:
                        self._log.debug(
                            "key {} on item {} is null "
                            "or empty: setting from item {.filepath}",
                            f,
                            displayable_path(objs[0].path),
                            o,
                        )
                        setattr(objs[0], f, value)
                        objs[0].store()
                        break
        return objs

    def _merge_albums(self, objs):
        """Merge Album objs by copying missing items from albums in the tail
        to the head album.

        Return same number of albums, with the head album modified."""
        ids = [i.mb_trackid for i in objs[0].items()]
        for o in objs[1:]:
            for i in o.items():
                if i.mb_trackid not in ids:
                    missing = Item.from_path(i.path)
                    missing.album_id = objs[0].id
                    missing.add(i._db)
                    self._log.debug(
                        "item {} missing from album {}:"
                        " merging from {.filepath} into {}",
                        missing,
                        objs[0],
                        o,
                        displayable_path(missing.destination()),
                    )
                    missing.move(operation=MoveOperation.COPY)
        return objs

    def _merge(self, objs):
        """Merge duplicate items. See ``_merge_items`` and ``_merge_albums``
        for the relevant strategies.
        """
        kind = Item if all(isinstance(o, Item) for o in objs) else Album
        if kind is Item:
            objs = self._merge_items(objs)
        else:
            objs = self._merge_albums(objs)
        return objs

    def _duplicates(self, objs, keys, full, strict, tiebreak, merge):
        """Generate triples of keys, duplicate counts, and constituent objects."""
        offset = 0 if full else 1
        for k, objs in self._group_by(objs, keys, strict).items():
            if len(objs) > 1:
                objs = self._order(objs, tiebreak)
                if merge:
                    objs = self._merge(objs)
                yield (k, len(objs) - offset, objs[offset:])



from __future__ import annotations

from typing import TYPE_CHECKING, NamedTuple

import beets

from .color import (
    ESC_TEXT_REGEX,
    RESET_COLOR,
    color_len,
    color_split,
    uncolorize,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator


class Side(NamedTuple):
    """A labeled segment of a two-column layout row with optional fixed width.

    Holds prefix, content, and suffix strings that together form one side of
    a formatted row. Width measurements account for ANSI color codes, which
    do not contribute to visible character count.
    """

    prefix: str
    contents: str
    suffix: str
    width: int = -1

    @property
    def rendered(self) -> str:
        """Assemble the full display string by joining prefix, contents, and suffix."""
        return f"{self.prefix}{self.contents}{self.suffix}"

    @property
    def prefix_width(self) -> int:
        """Visible character width of the prefix, excluding color codes."""
        return color_len(self.prefix)

    @property
    def suffix_width(self) -> int:
        """Visible character width of the suffix, excluding color codes."""
        return color_len(self.suffix)

    @property
    def rendered_width(self) -> int:
        """Visible character width of the fully assembled string."""
        return color_len(self.rendered)


def indent(count: int) -> str:
    """Returns a string with `count` many spaces."""
    return " " * count


def split_into_lines(string: str, first_width: int, width: int) -> list[str]:
    """Split string into a list of substrings at whitespace.

    The first substring has a length not longer than `first_width`, and the rest
    of substrings have a length not longer than `width`.

    `string` may contain ANSI codes at word borders.
    """
    words = []

    if uncolorize(string) == string:
        # No colors in string
        words = string.split()
    else:
        # Use a regex to find escapes and the text within them.
        for m in ESC_TEXT_REGEX.finditer(string):
            # m contains four groups:
            # pretext - any text before escape sequence
            # esc - intitial escape sequence
            # text - text, no escape sequence, may contain spaces
            # reset - ASCII colour reset
            space_before_text = False
            if m.group("pretext") != "":
                # Some pretext found, let's handle it
                # Add any words in the pretext
                words += m.group("pretext").split()
                if m.group("pretext")[-1] == " ":
                    # Pretext ended on a space
                    space_before_text = True
                else:
                    # Pretext ended mid-word, ensure next word
                    pass
            else:
                # pretext empty, treat as if there is a space before
                space_before_text = True
            if m.group("text")[0] == " ":
                # First character of the text is a space
                space_before_text = True
            # Now, handle the words in the main text:
            raw_words = m.group("text").split()
            if space_before_text:
                # Colorize each word with pre/post escapes
                # Reconstruct colored words
                words += [
                    f"{m['esc']}{raw_word}{RESET_COLOR}"
                    for raw_word in raw_words
                ]
            elif raw_words:
                # Pretext stops mid-word
                if m.group("esc") != RESET_COLOR:
                    # Add the rest of the current word, with a reset after it
                    words[-1] += f"{m['esc']}{raw_words[0]}{RESET_COLOR}"
                    # Add the subsequent colored words:
                    words += [
                        f"{m['esc']}{raw_word}{RESET_COLOR}"
                        for raw_word in raw_words[1:]
                    ]
                else:
                    # Caught a mid-word escape sequence
                    words[-1] += raw_words[0]
                    words += raw_words[1:]
            if (
                m.group("text")[-1] != " "
                and m.group("posttext") != ""
                and m.group("posttext")[0] != " "
            ):
                # reset falls mid-word
                post_text = m.group("posttext").split()
                words[-1] += post_text[0]
                words += post_text[1:]
            else:
                # Add any words after escape sequence
                words += m.group("posttext").split()
    result: list[str] = []
    next_substr = ""
    # Iterate over all words.
    previous_fit = False
    for i in range(len(words)):
        if i == 0:
            pot_substr = words[i]
        else:
            # (optimistically) add the next word to check the fit
            pot_substr = " ".join([next_substr, words[i]])
        # Find out if the pot(ential)_substr fits into the next substring.
        fits_first = len(result) == 0 and color_len(pot_substr) <= first_width
        fits_middle = len(result) != 0 and color_len(pot_substr) <= width
        if fits_first or fits_middle:
            # Fitted(!) let's try and add another word before appending
            next_substr = pot_substr
            previous_fit = True
        elif not fits_first and not fits_middle and previous_fit:
            # Extra word didn't fit, append what we have
            result.append(next_substr)
            next_substr = words[i]
            previous_fit = color_len(next_substr) <= width
        else:
            # Didn't fit anywhere
            if uncolorize(pot_substr) == pot_substr:
                # Simple uncolored string, append a cropped word
                if len(result) == 0:
                    # Crop word by the first_width for the first line
                    result.append(pot_substr[:first_width])
                    # add rest of word to next line
                    next_substr = pot_substr[first_width:]
                else:
                    result.append(pot_substr[:width])
                    next_substr = pot_substr[width:]
            else:
                # Colored strings
                if len(result) == 0:
                    this_line, next_line = color_split(pot_substr, first_width)
                    result.append(this_line)
                    next_substr = next_line
                else:
                    this_line, next_line = color_split(pot_substr, width)
                    result.append(this_line)
                    next_substr = next_line
            previous_fit = color_len(next_substr) <= width

    # We finished constructing the substrings, but the last substring
    # has not yet been added to the result.
    result.append(next_substr)
    return result


def get_column_layout(
    indent_str: str, left: Side, right: Side, max_width: int, separator: str
) -> Iterator[str]:
    """Print left & right data, with separator inbetween
    'left' and 'right' have a structure of:
    {'prefix':u'','contents':u'','suffix':u'','width':0}
    In a column layout the printing will be:
    {indent_str}{lhs0}{separator}{rhs0}
            {lhs1 / padding }{rhs1}
            ...
    The first line of each column (i.e. {lhs0} or {rhs0}) is:
    {prefix}{part of contents}{suffix}
    With subsequent lines (i.e. {lhs1}, {rhs1} onwards) being the
    rest of contents, wrapped if the width would be otherwise exceeded.
    """
    if left.width == -1 or right.width == -1:
        # If widths have not been defined, set to share space.
        width = (max_width - len(indent_str) - len(separator)) // 2
        left = left._replace(width=width)
        right = right._replace(width=width)
    # On the first line, account for suffix as well as prefix
    left_width_without_prefix = left.width - left.prefix_width
    left_split = split_into_lines(
        left.contents,
        left_width_without_prefix - left.suffix_width,
        left_width_without_prefix,
    )

    right_width_without_prefix = right.width - right.prefix_width
    right_split = split_into_lines(
        right.contents,
        right_width_without_prefix - right.suffix_width,
        right_width_without_prefix,
    )

    max_line_count = max(len(left_split), len(right_split))

    out = ""
    for i in range(max_line_count):
        # indentation
        out += indent_str

        # Prefix or indent_str for line
        if i == 0:
            out += left.prefix
        else:
            out += indent(left.prefix_width)

        # Line i of left hand side contents.
        if i < len(left_split):
            out += left_split[i]
            left_part_len = color_len(left_split[i])
        else:
            left_part_len = 0

        # Padding until end of column.
        # Note: differs from original
        # column calcs in not -1 afterwards for space
        # in track number as that is included in 'prefix'
        padding = left.width - left.prefix_width - left_part_len

        # Remove some padding on the first line to display
        # length
        if i == 0:
            padding -= left.suffix_width

        out += indent(padding)

        if i == 0:
            out += left.suffix

        # Separator between columns.
        if i == 0:
            out += separator
        else:
            out += indent(len(separator))

        # Right prefix, contents, padding, suffix
        if i == 0:
            out += right.prefix
        else:
            out += indent(right.prefix_width)

        # Line i of right hand side.
        if i < len(right_split):
            out += right_split[i]
            right_part_len = color_len(right_split[i])
        else:
            right_part_len = 0

        # Padding until end of column
        padding = right.width - right.prefix_width - right_part_len
        # Remove some padding on the first line to display
        # length
        if i == 0:
            padding -= right.suffix_width
        out += indent(padding)
        # Length in first line
        if i == 0:
            out += right.suffix

        # Linebreak, except in the last line.
        if i < max_line_count - 1:
            out += "\n"

    # Constructed all of the columns, now print
    yield out


def get_newline_layout(
    indent_str: str, left: Side, right: Side, max_width: int, separator: str
) -> Iterator[str]:
    """Prints using a newline separator between left & right if
    they go over their allocated widths. The datastructures are
    shared with the column layout. In contrast to the column layout,
    the prefix and suffix are printed at the beginning and end of
    the contents. If no wrapping is required (i.e. everything fits) the
    first line will look exactly the same as the column layout:
    {indent}{lhs0}{separator}{rhs0}
    However if this would go over the width given, the layout now becomes:
    {indent}{lhs0}
    {indent}{separator}{rhs0}
    If {lhs0} would go over the maximum width, the subsequent lines are
    indented a second time for ease of reading.
    """
    width_without_prefix = max_width - len(indent_str)
    width_without_double_prefix = max_width - 2 * len(indent_str)
    # On lower lines we will double the indent for clarity
    left_split = split_into_lines(
        left.rendered, width_without_prefix, width_without_double_prefix
    )
    # Repeat calculations for rhs, including separator on first line
    right_split = split_into_lines(
        right.rendered,
        width_without_prefix - len(separator),
        width_without_double_prefix,
    )
    for i, line in enumerate(left_split):
        if i == 0:
            yield f"{indent_str}{line}"
        elif line != "":
            # Ignore empty lines
            yield f"{indent_str * 2}{line}"
    for i, line in enumerate(right_split):
        if i == 0:
            yield f"{indent_str}{separator}{line}"
        elif line != "":
            yield f"{indent_str * 2}{line}"


def get_layout_method() -> Callable[[str, Side, Side, int, str], Iterator[str]]:
    return beets.config["ui"]["import"]["layout"].as_choice(
        {"column": get_column_layout, "newline": get_newline_layout}
    )


def get_layout_lines(
    indent_str: str, left: Side, right: Side, max_width: int
) -> Iterator[str]:
    # No right hand information, so we don't need a separator.
    separator = "" if right.rendered == "" else " -> "
    first_line_no_wrap = (
        f"{indent_str}{left.rendered}{separator}{right.rendered}"
    )
    if color_len(first_line_no_wrap) < max_width:
        # Everything fits, print out line.
        yield first_line_no_wrap
    else:
        layout_method = get_layout_method()
        yield from layout_method(indent_str, left, right, max_width, separator)



"""Updates a Sonos library whenever the beets library is changed.
This is based on the Kodi Update plugin.
"""

import soco

from beets.plugins import BeetsPlugin


class SonosUpdate(BeetsPlugin):
    def __init__(self):
        super().__init__()
        self.register_listener("database_change", self.listen_for_db_change)

    def listen_for_db_change(self, lib, model):
        """Listens for beets db change and register the update"""
        self.register_listener("cli_exit", self.update)

    def update(self, lib):
        """When the client exists try to send refresh request to a Sonos
        controller.
        """
        self._log.info("Requesting a Sonos library update...")

        device = soco.discovery.any_soco()

        if device:
            device.music_library.start_library_update()
        else:
            self._log.warning("Could not find a Sonos device.")
            return

        self._log.info("Sonos update triggered")



"""Fetch various AcousticBrainz metadata using MBID."""

from collections import defaultdict
from typing import ClassVar

import requests

from beets import plugins, ui
from beets.dbcore import types
from beets.exceptions import UserError

LEVELS = ["/low-level", "/high-level"]
ABSCHEME = {
    "highlevel": {
        "danceability": {"all": {"danceable": "danceable"}},
        "gender": {"value": "gender"},
        "genre_rosamerica": {"value": "genre_rosamerica"},
        "mood_acoustic": {"all": {"acoustic": "mood_acoustic"}},
        "mood_aggressive": {"all": {"aggressive": "mood_aggressive"}},
        "mood_electronic": {"all": {"electronic": "mood_electronic"}},
        "mood_happy": {"all": {"happy": "mood_happy"}},
        "mood_party": {"all": {"party": "mood_party"}},
        "mood_relaxed": {"all": {"relaxed": "mood_relaxed"}},
        "mood_sad": {"all": {"sad": "mood_sad"}},
        "moods_mirex": {"value": "moods_mirex"},
        "ismir04_rhythm": {"value": "rhythm"},
        "tonal_atonal": {"all": {"tonal": "tonal"}},
        "timbre": {"value": "timbre"},
        "voice_instrumental": {"value": "voice_instrumental"},
    },
    "lowlevel": {"average_loudness": "average_loudness"},
    "rhythm": {"bpm": "bpm"},
    "tonal": {
        "chords_changes_rate": "chords_changes_rate",
        "chords_key": "chords_key",
        "chords_number_rate": "chords_number_rate",
        "chords_scale": "chords_scale",
        "key_key": ("initial_key", 0),
        "key_scale": ("initial_key", 1),
        "key_strength": "key_strength",
    },
}


class AcousticPlugin(plugins.BeetsPlugin):
    item_types: ClassVar[dict[str, types.Type]] = {
        "average_loudness": types.Float(6),
        "chords_changes_rate": types.Float(6),
        "chords_key": types.STRING,
        "chords_number_rate": types.Float(6),
        "chords_scale": types.STRING,
        "danceable": types.Float(6),
        "gender": types.STRING,
        "genre_rosamerica": types.STRING,
        "initial_key": types.STRING,
        "key_strength": types.Float(6),
        "mood_acoustic": types.Float(6),
        "mood_aggressive": types.Float(6),
        "mood_electronic": types.Float(6),
        "mood_happy": types.Float(6),
        "mood_party": types.Float(6),
        "mood_relaxed": types.Float(6),
        "mood_sad": types.Float(6),
        "moods_mirex": types.STRING,
        "rhythm": types.Float(6),
        "timbre": types.STRING,
        "tonal": types.Float(6),
        "voice_instrumental": types.STRING,
    }

    def __init__(self):
        super().__init__()

        self._log.warning("This plugin is deprecated.")

        self.config.add(
            {"auto": True, "force": False, "tags": [], "base_url": ""}
        )

        self.base_url = self.config["base_url"].as_str()
        if self.base_url:
            if not self.base_url.startswith("http"):
                raise UserError(
                    "AcousticBrainz server base URL must start "
                    "with an HTTP scheme"
                )
            if self.base_url[-1] != "/":
                self.base_url = f"{self.base_url}/"

        if self.config["auto"]:
            self.register_listener("import_task_files", self.import_task_files)

    def commands(self):
        cmd = ui.Subcommand(
            "acousticbrainz", help="fetch metadata from AcousticBrainz"
        )
        cmd.parser.add_option(
            "-f",
            "--force",
            dest="force_refetch",
            action="store_true",
            default=False,
            help="re-download data when already present",
        )

        def func(lib, opts, args):
            items = lib.items(args)
            self._fetch_info(
                items,
                ui.should_write(),
                opts.force_refetch or self.config["force"],
            )

        cmd.func = func
        return [cmd]

    def import_task_files(self, session, task):
        """Function is called upon beet import."""
        self._fetch_info(task.imported_items(), False, True)

    def _get_data(self, mbid):
        if not self.base_url:
            raise UserError(
                "This plugin is deprecated since AcousticBrainz has shut "
                "down. See the base_url configuration option."
            )
        data = {}
        for url in _generate_urls(self.base_url, mbid):
            self._log.debug("fetching URL: {}", url)

            try:
                res = requests.get(url, timeout=10)
            except requests.RequestException as exc:
                self._log.info("request error: {}", exc)
                return {}

            if res.status_code == 404:
                self._log.info("recording ID {} not found", mbid)
                return {}

            try:
                data.update(res.json())
            except ValueError:
                self._log.debug("Invalid Response: {.text}", res)
                return {}

        return data

    def _fetch_info(self, items, write, force):
        """Fetch additional information from AcousticBrainz for the `item`s."""
        tags = self.config["tags"].as_str_seq()
        for item in items:
            # If we're not forcing re-downloading for all tracks, check
            # whether the data is already present. We use one
            # representative field name to check for previously fetched
            # data.
            if not force:
                mood_str = item.get("mood_acoustic", "")
                if mood_str:
                    self._log.info("data already present for: {}", item)
                    continue

            # We can only fetch data for tracks with MBIDs.
            if not item.mb_trackid:
                continue

            self._log.info("getting data for: {}", item)
            data = self._get_data(item.mb_trackid)
            if data:
                for attr, val in self._map_data_to_scheme(data, ABSCHEME):
                    if not tags or attr in tags:
                        self._log.debug(
                            "attribute {} of {} set to {}", attr, item, val
                        )
                        setattr(item, attr, val)
                    else:
                        self._log.debug(
                            "skipping attribute {} of {}"
                            " (value {}) due to config",
                            attr,
                            item,
                            val,
                        )
                item.store()
                if write:
                    item.try_write()

    def _map_data_to_scheme(self, data, scheme):
        """Given `data` as a structure of nested dictionaries, and
        `scheme` as a structure of nested dictionaries , `yield` tuples
        `(attr, val)` where `attr` and `val` are corresponding leaf
        nodes in `scheme` and `data`.

        As its name indicates, `scheme` defines how the data is structured,
        so this function tries to find leaf nodes in `data` that correspond
        to the leafs nodes of `scheme`, and not the other way around.
        Leaf nodes of `data` that do not exist in the `scheme` do not matter.
        If a leaf node of `scheme` is not present in `data`,
        no value is yielded for that attribute and a simple warning is issued.

        Finally, to account for attributes of which the value is split between
        several leaf nodes in `data`, leaf nodes of `scheme` can be tuples
        `(attr, order)` where `attr` is the attribute to which the leaf node
        belongs, and `order` is the place at which it should appear in the
        value. The different `value`s belonging to the same `attr` are simply
        joined with `' '`. This is hardcoded and not very flexible, but it gets
        the job done.

        For example:

        >>> scheme = {
            'key1': 'attribute',
            'key group': {
                'subkey1': 'subattribute',
                'subkey2': ('composite attribute', 0)
            },
            'key2': ('composite attribute', 1)
        }
        >>> data = {
            'key1': 'value',
            'key group': {
                'subkey1': 'subvalue',
                'subkey2': 'part 1 of composite attr'
            },
            'key2': 'part 2'
        }
        >>> print(list(_map_data_to_scheme(data, scheme)))
        [('subattribute', 'subvalue'),
         ('attribute', 'value'),
         ('composite attribute', 'part 1 of composite attr part 2')]
        """
        # First, we traverse `scheme` and `data`, `yield`ing all the non
        # composites attributes straight away and populating the dictionary
        # `composites` with the composite attributes.

        # When we are finished traversing `scheme`, `composites` should
        # map each composite attribute to an ordered list of the values
        # belonging to the attribute, for example:
        # `composites = {'initial_key': ['B', 'minor']}`.

        # The recursive traversal.
        composites = defaultdict(list)
        yield from self._data_to_scheme_child(data, scheme, composites)

        # When composites has been populated, yield the composite attributes
        # by joining their parts.
        for composite_attr, value_parts in composites.items():
            yield composite_attr, " ".join(value_parts)

    def _data_to_scheme_child(self, subdata, subscheme, composites):
        """The recursive business logic of :meth:`_map_data_to_scheme`:
        Traverse two structures of nested dictionaries in parallel and `yield`
        tuples of corresponding leaf nodes.

        If a leaf node belongs to a composite attribute (is a `tuple`),
        populate `composites` rather than yielding straight away.
        All the child functions for a single traversal share the same
        `composites` instance, which is passed along.
        """
        for k, v in subscheme.items():
            if k in subdata:
                if isinstance(v, dict):
                    yield from self._data_to_scheme_child(
                        subdata[k], v, composites
                    )
                elif isinstance(v, tuple):
                    composite_attribute, part_number = v
                    attribute_parts = composites[composite_attribute]
                    # Parts are not guaranteed to be inserted in order
                    while len(attribute_parts) <= part_number:
                        attribute_parts.append("")
                    attribute_parts[part_number] = subdata[k]
                else:
                    yield v, subdata[k]
            else:
                self._log.warning(
                    "Acousticbrainz did not provide info about {}", k
                )
                self._log.debug(
                    "Data {} could not be mapped to scheme {} "
                    "because key {} was not found",
                    subdata,
                    v,
                    k,
                )


def _generate_urls(base_url, mbid):
    """Generates AcousticBrainz end point urls for given `mbid`."""
    for level in LEVELS:
        yield f"{base_url}{mbid}{level}"
